# -*- coding: utf-8 -*-
"""
Module to generate surrogates of a spike train by randomising its spike times
in different ways (see :cite:`surrogates-Gerstein2004_203`,
:cite:`surrogates-Louis2010_127`, and :cite:`surrogates-Louis2010_359`).
Different methods destroy different features of the original data.


Main function
-------------
.. autosummary::
    :toctree: _toctree/spike_train_surrogates

    surrogates


Surrogate types
---------------
.. autosummary::
    :toctree: _toctree/spike_train_surrogates

    JointISI
    dither_spikes
    randomise_spikes
    shuffle_isis
    dither_spike_train
    jitter_spikes
    bin_shuffling
    trial_shifting

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import random
import warnings
import copy

import neo
import numpy as np
import quantities as pq
from scipy.ndimage import gaussian_filter

from elephant.statistics import isi
import elephant.conversion as conv
from elephant.utils import deprecated_alias

__all__ = [
    "dither_spikes",
    "randomise_spikes",
    "shuffle_isis",
    "dither_spike_train",
    "jitter_spikes",
    "bin_shuffling",
    "JointISI",
    "trial_shifting",
    "surrogates"
]

# List of all available surrogate methods
SURR_METHODS = ('dither_spike_train', 'dither_spikes', 'jitter_spikes',
                'randomise_spikes', 'shuffle_isis', 'joint_isi_dithering',
                'dither_spikes_with_refractory_period', 'trial_shifting',
                'bin_shuffling', 'isi_dithering')


def _dither_spikes_with_refractory_period(spiketrain, dither, n_surrogates,
                                          refractory_period):
    units = spiketrain.units
    t_start = spiketrain.t_start.rescale(units).magnitude
    t_stop = spiketrain.t_stop.rescale(units).magnitude

    dither = dither.rescale(units).magnitude
    refractory_period = refractory_period.rescale(units).magnitude
    # The initially guesses refractory period is compared to the minimal ISI.
    # The smaller value is taken as the refractory to calculate with.
    refractory_period = np.min(np.diff(spiketrain.magnitude),
                               initial=refractory_period)

    dithered_spiketrains = []
    for _ in range(n_surrogates):
        dithered_st = np.copy(spiketrain.magnitude)
        random_ordered_ids = np.arange(len(spiketrain))
        np.random.shuffle(random_ordered_ids)

        for random_id in random_ordered_ids:
            spike = dithered_st[random_id]
            prev_spike = dithered_st[random_id - 1] \
                if random_id > 0 \
                else t_start - refractory_period
            # subtract refractory period so that the first spike can move up
            # to t_start
            next_spike = dithered_st[random_id + 1] \
                if random_id < len(spiketrain) - 1 \
                else t_stop + refractory_period
            # add refractory period so that the last spike can move up
            # to t_stop

            # Dither range in the direction to the previous spike
            prev_dither = min(dither, spike - prev_spike - refractory_period)
            # Dither range in the direction to the next spike
            next_dither = min(dither, next_spike - spike - refractory_period)

            dt = (prev_dither + next_dither) * random.random() - prev_dither
            dithered_st[random_id] += dt

        dithered_spiketrains.append(dithered_st)

    dithered_spiketrains = np.array(dithered_spiketrains) * units

    return dithered_spiketrains


@deprecated_alias(n='n_surrogates')
def dither_spikes(spiketrain, dither, n_surrogates=1, decimals=None,
                  edges=True, refractory_period=None):
    """
    Generates surrogates of a spike train by spike dithering.

    The surrogates are obtained by uniformly dithering times around the
    original position. The dithering is performed independently for each
    surrogate.

    The surrogates retain the `spiketrain.t_start` and `spiketrain.t_stop`.
    Spikes moved beyond this range are lost or moved to the range's ends,
    depending on the parameter `edges`.

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        The spike train from which to generate the surrogates.
    dither : pq.Quantity
        Amount of dithering. A spike at time `t` is placed randomly within
        `(t-dither, t+dither)`.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1
    decimals : int or None, optional
        Number of decimal points for every spike time in the surrogates at a
        millisecond level.
        If None, machine precision is used.
        Default: None
    edges : bool, optional
        For surrogate spikes falling outside the range
        `[spiketrain.t_start, spiketrain.t_stop)`, whether to drop them out
        (for `edges = True`) or set them to the range's closest end
        (for `edges = False`).
        Default: True
    refractory_period : pq.Quantity or None, optional
        The dither range of each spike is adjusted such that the spike can not
        fall into the `refractory_period` of the previous or next spike.
        To account this, the refractory period is estimated as the smallest ISI
        of the spike train. The given argument `refractory_period` here is thus
        an initial estimation.
        Note, that with this option a spike cannot "jump" over the previous or
        next spike as it is normally possible.
        If set to None, no refractoriness is in dithering.
        Default: None

    Returns
    -------
    list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        randomly dithering its spikes. The range of the surrogate spike trains
        is the same as of `spiketrain`.

    Examples
    --------
    >>> import quantities as pq
    >>> import neo
    ...
    >>> st = neo.SpikeTrain([100, 250, 600, 800] * pq.ms, t_stop=1 * pq.s)
    >>> print(dither_spikes(st, dither = 20 * pq.ms))  # doctest: +SKIP
    [<SpikeTrain(array([  96.53801903,  248.57047376,  601.48865767,
     815.67209811]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(dither_spikes(st, dither = 20 * pq.ms, n_surrogates=2))
    [<SpikeTrain(array([ 104.24942044,  246.0317873 ,  584.55938657,
        818.84446913]) * ms, [0.0 ms, 1000.0 ms])>,
     <SpikeTrain(array([ 111.36693058,  235.15750163,  618.87388515,
        786.1807108 ]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(dither_spikes(st, dither = 20 * pq.ms,
                            decimals=0))  # doctest: +SKIP
    [<SpikeTrain(array([  81.,  242.,  595.,  799.]) * ms,
        [0.0 ms, 1000.0 ms])>]

    """
    if len(spiketrain) == 0:
        # return the empty spiketrain n times
        return [spiketrain.copy() for _ in range(n_surrogates)]

    units = spiketrain.units
    t_start = spiketrain.t_start.rescale(units).magnitude
    t_stop = spiketrain.t_stop.rescale(units).magnitude

    if refractory_period is None or refractory_period == 0:
        # Main: generate the surrogates
        dither = dither.rescale(units).magnitude
        dithered_spiketrains = \
            spiketrain.magnitude.reshape((1, len(spiketrain))) \
            + 2 * dither * np.random.random_sample(
                (n_surrogates, len(spiketrain))) - dither
        dithered_spiketrains.sort(axis=1)

        if edges:
            # Leave out all spikes outside
            # [spiketrain.t_start, spiketrain.t_stop]
            dithered_spiketrains = \
                [train[
                    np.all([t_start < train, train < t_stop], axis=0)]
                 for train in dithered_spiketrains]
        else:
            # Move all spikes outside
            # [spiketrain.t_start, spiketrain.t_stop] to the range's ends
            dithered_spiketrains = np.minimum(
                np.maximum(dithered_spiketrains, t_start),
                t_stop)

        dithered_spiketrains = dithered_spiketrains * units

    elif isinstance(refractory_period, pq.Quantity):
        dithered_spiketrains = _dither_spikes_with_refractory_period(
            spiketrain, dither, n_surrogates, refractory_period)
    else:
        raise ValueError("refractory_period must be of type pq.Quantity")

    # Round the surrogate data to decimal position, if requested
    if decimals is not None:
        dithered_spiketrains = \
            dithered_spiketrains.rescale(pq.ms).round(decimals).rescale(units)

    # Return the surrogates as list of neo.SpikeTrain
    return [neo.SpikeTrain(train, t_start=t_start, t_stop=t_stop,
                           sampling_rate=spiketrain.sampling_rate)
            for train in dithered_spiketrains]


@deprecated_alias(n='n_surrogates')
def randomise_spikes(spiketrain, n_surrogates=1, decimals=None):
    """
    Generates surrogates of a spike train by spike time randomization.

    The surrogates are obtained by keeping the spike count of the original
    `spiketrain`, but placing the spikes randomly in the interval
    `[spiketrain.t_start, spiketrain.t_stop]`. The generated independent
    `neo.SpikeTrain` objects follow  Poisson statistics (exponentially
    distributed inter-spike intervals).

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        The spike train from which to generate the surrogates.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1
    decimals : int or None, optional
        Number of decimal points for every spike time in the surrogates.
        If None, machine precision is used.
        Default: None

    Returns
    -------
    list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        randomly distributing its spikes in the interval
        `[spiketrain.t_start, spiketrain.t_stop]`.

    Examples
    --------
    >>> import quantities as pq
    >>> import neo
    ...
    >>> st = neo.SpikeTrain([100, 250, 600, 800] * pq.ms, t_stop=1 * pq.s)
    >>> print(randomise_spikes(st))  # doctest: +SKIP
        [<SpikeTrain(array([ 131.23574603,  262.05062963,  549.84371387,
                            940.80503832]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(randomise_spikes(st, n_surrogates=2))  # doctest: +SKIP
        [<SpikeTrain(array([  84.53274955,  431.54011743,  733.09605806,
              852.32426583]) * ms, [0.0 ms, 1000.0 ms])>,
         <SpikeTrain(array([ 197.74596726,  528.93517359,  567.44599968,
              775.97843799]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(randomise_spikes(st, decimals=0))  # doctest: +SKIP
        [<SpikeTrain(array([  29.,  667.,  720.,  774.]) * ms,
              [0.0 ms, 1000.0 ms])>]

    """
    # Create surrogate spike trains as rows of a Quantity array
    sts = ((spiketrain.t_stop - spiketrain.t_start) *
           np.random.random(size=(n_surrogates, len(spiketrain))) +
           spiketrain.t_start).rescale(spiketrain.units)

    # Round the surrogate data to decimal position, if requested
    if decimals is not None:
        sts = sts.round(decimals)

    # Convert the Quantity array to a list of SpikeTrains, and return them
    return [neo.SpikeTrain(np.sort(st), t_start=spiketrain.t_start,
                           t_stop=spiketrain.t_stop,
                           sampling_rate=spiketrain.sampling_rate)
            for st in sts]


@deprecated_alias(n='n_surrogates')
def shuffle_isis(spiketrain, n_surrogates=1, decimals=None):
    """
    Generates surrogates of a spike train by inter-spike-interval (ISI)
    shuffling.

    The surrogates are obtained by randomly sorting the ISIs of the given input
    `spiketrain`. This generates independent `neo.SpikeTrain` object(s) with
    same ISI distribution and spike count as in `spiketrain`, while
    destroying temporal dependencies and firing rate profile.

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        The spike train from which to generate the surrogates.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1
    decimals : int or None, optional
        Number of decimal points for every spike time in the surrogates.
        If None, machine precision is used.
        Default: None

    Returns
    -------
    list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        random ISI shuffling. The time range of the surrogate spike trains is
        the same as in `spiketrain`.

    Examples
    --------
    >>> import quantities as pq
    >>> import neo
    ...
    >>> st = neo.SpikeTrain([100, 250, 600, 800] * pq.ms, t_stop=1 * pq.s)
    >>> print(shuffle_isis(st))  # doctest: +SKIP
        [<SpikeTrain(array([ 200.,  350.,  700.,  800.]) * ms,
                 [0.0 ms, 1000.0 ms])>]
    >>> print(shuffle_isis(st, n_surrogates=2))  # doctest: +SKIP
        [<SpikeTrain(array([ 100.,  300.,  450.,  800.]) * ms,
              [0.0 ms, 1000.0 ms])>,
         <SpikeTrain(array([ 200.,  350.,  700.,  800.]) * ms,
              [0.0 ms, 1000.0 ms])>]

    """
    if len(spiketrain) == 0:
        return [neo.SpikeTrain([] * spiketrain.units,
                               t_start=spiketrain.t_start,
                               t_stop=spiketrain.t_stop,
                               sampling_rate=spiketrain.sampling_rate)
                for _ in range(n_surrogates)]

    # A correct sorting is necessary, to calculate the ISIs
    spiketrain = spiketrain.copy()
    spiketrain.sort()
    isi0 = spiketrain[0] - spiketrain.t_start
    isis = np.hstack([isi0, isi(spiketrain)])

    # Round the isis to decimal position, if requested
    if decimals is not None:
        isis = isis.round(decimals)

    # Create list of surrogate spike trains by random ISI permutation
    sts = []
    for surrogate_id in range(n_surrogates):
        surr_times = np.cumsum(np.random.permutation(isis)) * \
            spiketrain.units + spiketrain.t_start
        sts.append(neo.SpikeTrain(
            surr_times, t_start=spiketrain.t_start,
            t_stop=spiketrain.t_stop,
            sampling_rate=spiketrain.sampling_rate))
    return sts


@deprecated_alias(n='n_surrogates')
def dither_spike_train(spiketrain, shift, n_surrogates=1, decimals=None,
                       edges=True):
    """
    Generates surrogates of a spike train by spike train shifting.

    The surrogates are obtained by shifting the whole spike train by a
    random amount of time (independent for each surrogate). Thus, ISIs and
    temporal correlations within the spike train are kept. For small shifts,
    the firing rate profile is also kept with reasonable accuracy.

    The surrogates retain the `spiketrain.t_start` and `spiketrain.t_stop`.
    Spikes moved beyond this range are lost or moved to the range's ends,
    depending on the parameter `edges`.

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        The spike train from which to generate the surrogates.
    shift : pq.Quantity
        Amount of shift. `spiketrain` is shifted by a random amount uniformly
        drawn from the range `(-shift, +shift)`.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1
    decimals : int or None, optional
        Number of decimal points for every spike time in the surrogates.
        If None, machine precision is used.
        Default: None
    edges : bool, optional
        For surrogate spikes falling outside the range `[spiketrain.t_start,
        spiketrain.t_stop)`, whether to drop them out (for `edges = True`) or
        set them to the range's closest end (for `edges = False`).
        Default: True

    Returns
    -------
    list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        randomly dithering the whole spike train. The time range of the
        surrogate spike trains is the same as in `spiketrain`.

    Examples
    --------
    >>> import quantities as pq
    >>> import neo
    ...
    >>> st = neo.SpikeTrain([100, 250, 600, 800] * pq.ms, t_stop=1 * pq.s)
    >>> print(dither_spike_train(st, shift = 20*pq.ms))  # doctest: +SKIP
    [<SpikeTrain(array([  96.53801903,  248.57047376,  601.48865767,
     815.67209811]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(dither_spike_train(st, shift = 20*pq.ms, n_surrogates=2))
    [<SpikeTrain(array([  92.89084054,  242.89084054,  592.89084054,
        792.89084054]) * ms, [0.0 ms, 1000.0 ms])>,
     <SpikeTrain(array([  84.61079043,  234.61079043,  584.61079043,
        784.61079043]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(dither_spike_train(st, shift = 20 * pq.ms,
                                 decimals=0))  # doctest: +SKIP
    [<SpikeTrain(array([  82.,  232.,  582.,  782.]) * ms,
        [0.0 ms, 1000.0 ms])>]

    """
    # Transform spiketrain into a Quantity object (needed for matrix algebra)
    data = spiketrain.view(pq.Quantity)

    # Main: generate the surrogates by spike train shifting
    surr = data.reshape(
        (1, len(data))) + 2 * shift * np.random.random_sample(
            (n_surrogates, 1)) - shift

    # Round the surrogate data to decimal position, if requested
    if decimals is not None:
        surr = surr.round(decimals)

    if edges is False:
        # Move all spikes outside [spiketrain.t_start, spiketrain.t_stop] to
        # the range's ends
        surr = np.minimum(np.maximum(surr.simplified.magnitude,
                                     spiketrain.t_start.simplified.magnitude),
                          spiketrain.t_stop.simplified.magnitude) * pq.s
    else:
        # Leave out all spikes outside [spiketrain.t_start, spiketrain.t_stop]
        tstart, tstop = spiketrain.t_start.simplified.magnitude, \
            spiketrain.t_stop.simplified.magnitude
        surr = [np.sort(s[np.all([s >= tstart, s < tstop], axis=0)]) * pq.s
                for s in surr.simplified.magnitude]

    # Return the surrogates as SpikeTrains
    return [neo.SpikeTrain(s, t_start=spiketrain.t_start,
                           t_stop=spiketrain.t_stop,
                           sampling_rate=spiketrain.sampling_rate
                           ).rescale(spiketrain.units)
            for s in surr]


@deprecated_alias(binsize='bin_size', n='n_surrogates')
def jitter_spikes(spiketrain, bin_size, n_surrogates=1):
    """
    Generates surrogates of a spike train by spike jittering.

    The surrogates are obtained by defining adjacent time bins spanning the
    `spiketrain` range, and randomly re-positioning (independently for each
    surrogate) each spike in the time bin it falls into.

    The surrogates retain the `spiketrain.t_start` and `spiketrain.t_stop`.
    Note that within each time bin the surrogate spike trains are locally
    Poissonian (the inter-spike-intervals are exponentially distributed).

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        The spike train from which to generate the surrogates.
    bin_size : pq.Quantity
        Size of the time bins within which to randomize the spike times.
        Note: the last bin lasts until `spiketrain.t_stop` and might have
        width different from `bin_size`.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1

    Returns
    -------
    list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        randomly replacing its spikes within bins of user-defined width. The
        time range of the surrogate spike trains is the same as in
        `spiketrain`.

    Examples
    --------
    >>> import quantities as pq
    >>> import neo
    ...
    >>> st = neo.SpikeTrain([80, 150, 320, 480] * pq.ms, t_stop=1 * pq.s)
    >>> print(jitter_spikes(st, bin_size=100 * pq.ms))  # doctest: +SKIP
    [<SpikeTrain(array([  98.82898293,  178.45805954,  346.93993867,
        461.34268507]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(jitter_spikes(st, bin_size=100 * pq.ms, n_surrogates=2))
    [<SpikeTrain(array([  97.15720041,  199.06945744,  397.51928207,
        402.40065162]) * ms, [0.0 ms, 1000.0 ms])>,
     <SpikeTrain(array([  80.74513157,  173.69371317,  338.05860962,
        495.48869981]) * ms, [0.0 ms, 1000.0 ms])>]
    >>> print(jitter_spikes(st, bin_size=100 * pq.ms))  # doctest: +SKIP
    [<SpikeTrain(array([  4.55064897e-01,   1.31927046e+02,   3.57846265e+02,
         4.69370604e+02]) * ms, [0.0 ms, 1000.0 ms])>]

    """
    # Define standard time unit; all time Quantities are converted to
    # scalars after being rescaled to this unit, to use the power of numpy
    std_unit = bin_size.units

    # Compute bin edges for the jittering procedure
    # !: the last bin arrives until spiketrain.t_stop and might have
    # size != bin_size
    start_dl = spiketrain.t_start.rescale(std_unit).magnitude
    stop_dl = spiketrain.t_stop.rescale(std_unit).magnitude

    bin_edges = start_dl + np.arange(start_dl, stop_dl, bin_size.magnitude)
    bin_edges = np.hstack([bin_edges, stop_dl])

    # Create n surrogates with spikes randomly placed in the interval (0,1)
    surr_poiss01 = np.random.random_sample((n_surrogates, len(spiketrain)))

    # Compute the bin id of each spike
    bin_ids = np.array(
        (spiketrain.view(pq.Quantity) /
         bin_size).rescale(pq.dimensionless).magnitude, dtype=int)

    # Compute the size of each time bin (as a numpy array)
    bin_sizes_dl = np.diff(bin_edges)

    # For each spike compute its offset (the left end of the bin it falls
    # into) and the size of the bin it falls into
    offsets = start_dl + np.array([bin_edges[bin_id] for bin_id in bin_ids])
    dilats = np.array([bin_sizes_dl[bin_id] for bin_id in bin_ids])

    # Compute each surrogate by dilating and shifting each spike s in the
    # poisson 0-1 spike trains to dilat * s + offset. Attach time unit again
    surr = np.sort(surr_poiss01 * dilats + offsets, axis=1) * std_unit

    return [neo.SpikeTrain(s, t_start=spiketrain.t_start,
                           t_stop=spiketrain.t_stop,
                           sampling_rate=spiketrain.sampling_rate
                           ).rescale(spiketrain.units)
            for s in surr]


def bin_shuffling(
        spiketrain, max_displacement, bin_size=None, n_surrogates=1,
        sliding=False):
    """
    Bin shuffling surrogate generation.

    The function shuffles the entries of a binned spike train entries inside
    windows with a fixed maximal displacement. The windows are either exclusive
    or sliding.

    Parameters
    ----------
    spiketrain : conv.BinnedSpikeTrain or neo.SpikeTrain
        The binned spike train or a continuous time spike train
        to create surrogates of.
    max_displacement : int
        Number of bins that a single spike can be displaced.
    bin_size : pq.Quantity or None
        the bin size needs to be specified only if a not-binned spike train
        is passed to the method
    n_surrogates : int, optional
        Number of surrogates to create.
        Default: 1
    sliding : bool, optional
        If True, the window is slided bin by bin
        (only implemented for binned spike trains).
        Default: False

    Returns
    -------
    binned_surrogates : list of conv.BinnedSpikeTrain or list of neo.SpikeTrain
        Each entry of the list is a surrogate spike train either binned or in
        continuous time.
    """
    if isinstance(spiketrain, neo.SpikeTrain):
        if bin_size is None:
            raise ValueError(
                'If you want to create surrogates from neo.SpikeTrain objects,'
                'you need to specify the bin_size')
        if sliding:
            warnings.warn(
                'The sliding option is not implemented yet for bin shuffling'
                ' on continuos time spike trains. Results are given for'
                ' sliding=False.', UserWarning)
        return _continuous_time_bin_shuffling(
            spiketrain, max_displacement=max_displacement, bin_size=bin_size,
            n_surrogates=n_surrogates)

    displacement_window = 2 * max_displacement

    binned_spiketrain_bool = spiketrain.to_bool_array()[0]
    st_length = len(binned_spiketrain_bool)

    surrogate_spiketrains = []
    for surrogate_id in range(n_surrogates):
        surrogate_spiketrain = np.copy(binned_spiketrain_bool)
        if sliding:
            for window_position in range(st_length - displacement_window):
                # shuffling the binned spike train within the window
                np.random.shuffle(
                    surrogate_spiketrain[
                        window_position:
                        window_position + displacement_window])
        else:
            windows = st_length // displacement_window
            windows_remainder = st_length % displacement_window
            for window_position in range(windows):
                # shuffling the binned spike train within the window
                np.random.shuffle(
                    surrogate_spiketrain[
                        window_position * displacement_window:
                        (window_position + 1) * displacement_window])
            if windows_remainder != 0:
                np.random.shuffle(
                    surrogate_spiketrain[windows * displacement_window:])
        surrogate_spiketrain = surrogate_spiketrain.reshape((1, st_length))

        surrogate_spiketrains.append(
            conv.BinnedSpikeTrain(
                surrogate_spiketrain,
                bin_size=spiketrain.bin_size,
                t_start=spiketrain.t_start,
                t_stop=spiketrain.t_stop,
                tolerance=None))
    return surrogate_spiketrains


def _continuous_time_bin_shuffling(spiketrain, max_displacement, bin_size,
                                   n_surrogates=1):
    """

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
    max_displacement : int
        number of bins that a single spike can be displaced
    bin_size : pq.Quantity
    n_surrogates : int, optional
        Default : 1

    Returns
    -------
    list of neo.SpikeTrain
    """
    units = spiketrain.units
    bin_size = bin_size.rescale(units).item()
    t_start = spiketrain.t_start.item()
    t_stop = spiketrain.t_stop.item()
    spiketrain_shifted = spiketrain.magnitude - t_start
    displacement_window = 2 * max_displacement

    binned_duration = int((t_stop - t_start) // bin_size)

    bin_indices = (spiketrain_shifted // bin_size).astype(int)

    split_indices = np.searchsorted(
        bin_indices,
        np.arange(displacement_window, binned_duration, displacement_window))

    bin_indices = np.split(
        bin_indices,
        split_indices)

    surrogate_spiketrains = []
    for surrogate_id in range(n_surrogates):
        surrogate_bin_indices = np.empty(shape=len(bin_indices),
                                         dtype=np.ndarray)
        for i, bin_indices_slice in enumerate(bin_indices):
            window_start = i*displacement_window

            random_indices = np.random.permutation(displacement_window)
            surrogate_bin_indices[i] = \
                random_indices[bin_indices_slice - window_start] \
                + window_start

        surrogate_bin_indices = np.concatenate(surrogate_bin_indices)

        bin_remainders = bin_size * np.random.random(len(spiketrain))

        surrogate_spiketrain = \
            surrogate_bin_indices * bin_size + bin_remainders + t_start

        # ensure last and first spike being inside the boundaries
        surrogate_spiketrain = surrogate_spiketrain[
            np.all((surrogate_spiketrain > t_start,
                    surrogate_spiketrain < t_stop),
                   axis=0)]

        surrogate_spiketrain.sort()

        surrogate_spiketrain = neo.SpikeTrain(
            surrogate_spiketrain,
            units=units,
            t_start=t_start,
            t_stop=t_stop,
            copy=False,
        )

        surrogate_spiketrains.append(surrogate_spiketrain)
    return surrogate_spiketrains


class JointISI(object):
    r"""
    Joint-ISI dithering implementation, based on the ideas from
    :cite:`surrogates-Gerstein2004_203` and :cite:`surrogates-Louis2010_127`.

    The main function is :func:`JointISI.dithering`.

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        Input spiketrain to create surrogates of.
    dither : pq.Quantity, optional
        This quantity describes the maximum displacement of a spike, when
        `method` is 'window'. It is also used for the uniform dithering for
        the spikes, which are outside the regime in the Joint-ISI
        histogram, where Joint-ISI dithering is applicable.
        Default: 15. * pq.ms
    truncation_limit : pq.Quantity, optional
        The Joint-ISI distribution of :math:`(ISI_i, ISI_{i+1})` is defined
        within the range :math:`[0, \infty)`. Since this is computationally not
        feasible, the Joint-ISI distribution is truncated for high ISI.
        The Joint-ISI histogram is calculated for
        :math:`(ISI_i, ISI_{i+1})` from 0 to `truncation_limit`.
        Default: 100. * pq.ms
    n_bins : int, optional
        The size of the joint-ISI-distribution will be
        `n_bins*n_bins/2`.
        Default: 100
    sigma : pq.Quantity, optional
        The standard deviation of the Gaussian kernel, with which
        the data is convolved.
        Default: 2. * pq.ms
    alternate : bool, optional
        If True, then all even spikes are dithered followed
        by all odd spikes. Otherwise, the spikes are dithered in ascending
        order from the first to the last spike.
        Default: True
    use_sqrt : bool, optional
        If True, the joint-ISI histogram is preprocessed by
        applying a square root (following :cite:`surrogates-Gerstein2004_203`).
        Default: False
    method : {'fast', 'window'}, optional
        * 'fast': the spike can move in the whole range between the
            previous and subsequent spikes (computationally efficient).
        * 'window': the spike movement is limited to the parameter `dither`.

        Default: 'window'
    cutoff : bool, optional
        If True, then the filtering of the Joint-ISI histogram is
        limited on the lower side by the minimal ISI.
        This can be necessary, if in the data there is a certain refractory
        period, which will be destroyed by the convolution with the
        2d-Gaussian function.
        Default: True
    refractory_period : pq.Quantity, optional
        Defines the refractory period of the dithered `spiketrain` unless
        the smallest ISI of the `spiketrain` is lower than this value.
        Default: 4. * pq.ms
    isi_dithering : bool, optional
        If True, the Joint-ISI distribution is evaluated as the outer product
        of the ISI-distribution with itself. Thus, all serial correlations are
        destroyed.
        Default: False
    """

    # The min number of spikes, required for dithering.
    # Otherwise, the original spiketrain is copied N times.
    MIN_SPIKES = 3

    @deprecated_alias(num_bins='n_bins', refr_period='refractory_period')
    def __init__(self,
                 spiketrain,
                 dither=15. * pq.ms,
                 truncation_limit=100. * pq.ms,
                 n_bins=100,
                 sigma=2. * pq.ms,
                 alternate=True,
                 use_sqrt=False,
                 method='window',
                 cutoff=True,
                 refractory_period=4. * pq.ms,
                 isi_dithering=False):

        if not isinstance(spiketrain, neo.SpikeTrain):
            raise TypeError('spiketrain must be of type neo.SpikeTrain')

        # A correct sorting is necessary to calculate the ISIs
        spiketrain = spiketrain.copy()
        spiketrain.sort()
        self.spiketrain = spiketrain
        self.truncation_limit = self._get_magnitude(truncation_limit)
        self.n_bins = n_bins

        self.dither = self._get_magnitude(dither)

        self.sigma = self._get_magnitude(sigma)
        self.alternate = alternate

        if method not in ('fast', 'window'):
            raise ValueError("The method can either be 'fast' or 'window', "
                             "but not '{}'".format(method))
        self.method = method

        refractory_period = self._get_magnitude(refractory_period)
        if not self._too_less_spikes:
            minimal_isi = np.min(self.isi)
            refractory_period = min(refractory_period, minimal_isi)
        self.refractory_period = refractory_period

        self.cutoff = cutoff
        self.use_sqrt = use_sqrt
        self._jisih_cumulatives = None

        self._max_change_index = self._isi_to_index(self.dither)
        self._max_change_isi = self._index_to_isi(self._max_change_index)

        self.isi_dithering = isi_dithering

    @property
    def refr_period(self):
        warnings.warn("'.refr_period' is deprecated; use '.refractory_period'",
                      DeprecationWarning)
        return self.refractory_period

    @property
    def num_bins(self):
        warnings.warn("'.num_bins' is deprecated; use '.n_bins'",
                      DeprecationWarning)
        return self.n_bins

    def _get_magnitude(self, quantity):
        """
        Parameters
        ----------
        quantity : pq.Quantity or float

        Returns
        -------
        magnitude : float
            The magnitude of `quantity`, rescaled to the units of the input
            :attr:`spiketrain`.
        """
        if isinstance(quantity, pq.Quantity):
            return quantity.rescale(self._unit).magnitude
        return quantity

    @property
    def _too_less_spikes(self):
        """
        This is a check if the :attr:`spiketrain` has enough spikes to evaluate
        the joint-ISI histogram.

        Returns
        -------
        bool
            If True, the spike train is so sparse, that this algorithm can't be
            applied properly. Than in dithering() copies of the spiketrains are
            returned.
        """
        return len(self.spiketrain) < self.MIN_SPIKES

    @property
    def _unit(self):
        """
        The unit of the spiketrain. Thus, the unit of the output surrogates.
        """
        return self.spiketrain.units

    @property
    def isi(self):
        """
        The inter-spike intervals of the spiketrain.

        Returns
        -------
        np.ndarray or None
            An array of inter-spike intervals of the `spiketrain`.
            None, if not enough spikes in the `spiketrain`.
        """
        if self._too_less_spikes:
            return None
        return isi(self.spiketrain.magnitude)

    @property
    def bin_width(self):
        return self.truncation_limit / self.n_bins

    def _isi_to_index(self, inter_spike_interval):
        """
        A function that gives for each ISI the corresponding index in the
        Joint-ISI distribution.

        Parameters
        ----------
        inter_spike_interval : np.ndarray or float
            An array of ISIs or a single ISI.

        Returns
        -------
        indices : np.ndarray or int
            The corresponding indices/index for each ISI. When the input is an
            array, also the output is.
        """
        return np.floor(inter_spike_interval / self.bin_width).astype(int)

    def _index_to_isi(self, isi_index):
        """
        Maps `isi_index` back to the original ISI.

        Parameters
        ----------
        isi_index : np.ndarray or int
            The index of ISI.

        Returns
        -------
        np.ndarray or float
            The corresponding ISI(s) for each indices/index. When the input is
            an array, also the output is.
        """
        return (isi_index + 0.5) * self.bin_width

    def joint_isi_histogram(self):
        """
        Calculates a 2D histogram of :math:`(ISI_i, ISI_{i+1})` and applies
        square root or gaussian filtering if necessary.

        Returns
        -------
        joint_isi_histogram : np.ndarray or None
            A np.ndarray with shape `n_bins` x `n_bins` containing the
            joint-ISI histogram.
            None, if not enough spikes in the `spiketrain`.
        """
        if self._too_less_spikes:
            return None
        isis = self.isi
        if not self.isi_dithering:
            joint_isi_histogram = np.histogram2d(
                isis[:-1], isis[1:],
                bins=[self.n_bins, self.n_bins],
                range=[[0., self.truncation_limit],
                       [0., self.truncation_limit]])[0]
        else:
            isi_histogram = np.histogram(
                isis,
                bins=self.n_bins,
                range=[0., self.truncation_limit])[0]
            joint_isi_histogram = np.outer(isi_histogram, isi_histogram)

        if self.use_sqrt:
            joint_isi_histogram = np.sqrt(joint_isi_histogram)

        if self.sigma:
            if self.cutoff:
                start_index = self._isi_to_index(self.refractory_period)
                joint_isi_histogram[
                    start_index:, start_index:] = gaussian_filter(
                        joint_isi_histogram[start_index:, start_index:],
                        sigma=self.sigma / self.bin_width)
                joint_isi_histogram[:start_index, :] = 0
                joint_isi_histogram[:, :start_index] = 0
            else:
                joint_isi_histogram = gaussian_filter(
                    joint_isi_histogram, sigma=self.sigma / self.bin_width)
        return joint_isi_histogram

    @staticmethod
    def _normalize_cumulative_distribution(array):
        """
        This function normalizes the cut-off of a cumulative distribution
        function so that the first element is again zero and the last element
        is one.

        Parameters
        ----------
        array : np.ndarray
            A monotonously increasing array as a part of an unnormalized
            cumulative distribution function.

        Returns
        -------
        np.ndarray
            Monotonously increasing array from 0 to 1.
            If `array` does not contain all equal elements, a only-zeros array
            is returned.
        """
        if array[-1] - array[0] > 0.:
            return (array - array[0]) / (array[-1] - array[0])
        return np.zeros_like(array)

    def dithering(self, n_surrogates=1):
        """
        Implementation of Joint-ISI-dithering for spike trains that pass the
        threshold of the dense rate. If not, a uniform dithered spike train is
        given back.

        Parameters
        ----------
        n_surrogates : int
            The number of dithered spiketrains to be returned.
            Default: 1

        Returns
        ----------
        dithered_sts : list of neo.SpikeTrain
            Spike trains, that are dithered versions of the given
            :attr:`spiketrain`.
        """
        if self._too_less_spikes:
            return [self.spiketrain] * n_surrogates

        # Checks, whether the preprocessing is already done.
        if self._jisih_cumulatives is None:
            self._determine_cumulative_functions()

        dithered_sts = []
        isi_to_dither = self.isi
        for _ in range(n_surrogates):
            dithered_isi = self._get_dithered_isi(isi_to_dither)

            dithered_st = self.spiketrain[0].magnitude + \
                np.r_[0., np.cumsum(dithered_isi)]
            sampling_rate = self.spiketrain.sampling_rate

            # Due to rounding errors, the last spike may be above t_stop.
            # If the case, this is set to t_stop.
            if dithered_st[-1] > self.spiketrain.t_stop:
                dithered_st[-1] = self.spiketrain.t_stop

            dithered_st = neo.SpikeTrain(dithered_st * self._unit,
                                         t_start=self.spiketrain.t_start,
                                         t_stop=self.spiketrain.t_stop,
                                         sampling_rate=sampling_rate)
            dithered_sts.append(dithered_st)
        return dithered_sts

    def _determine_cumulative_functions(self):
        rotated_jisih = np.rot90(self.joint_isi_histogram())

        if self.method == 'fast':
            self._jisih_cumulatives = []
            for double_index in range(self.n_bins):
                # Taking anti-diagonals of the original joint-ISI histogram
                diagonal = np.diagonal(
                    rotated_jisih, offset=-self.n_bins + double_index + 1)
                jisih_cum = self._normalize_cumulative_distribution(
                    np.r_[0., np.cumsum(diagonal)])
                self._jisih_cumulatives.append(jisih_cum)
            self._jisih_cumulatives = np.array(
                self._jisih_cumulatives, dtype=object)
        else:
            self._jisih_cumulatives = self._window_cumulatives(rotated_jisih)

    def _window_cumulatives(self, rotated_jisih):
        jisih_diag_cums = self._window_diagonal_cumulatives(rotated_jisih)
        jisih_cumulatives = np.zeros(
            (self.n_bins, self.n_bins,
             2 * self._max_change_index + 1))
        for curr_isi_id in range(self.n_bins):
            for next_isi_id in range(self.n_bins - curr_isi_id):
                double_index = next_isi_id + curr_isi_id
                cum_slice = jisih_diag_cums[
                    double_index,
                    curr_isi_id: curr_isi_id + 2 * self._max_change_index + 1]

                normalized_cum = self._normalize_cumulative_distribution(
                    cum_slice)
                jisih_cumulatives[curr_isi_id][next_isi_id] = normalized_cum
        return jisih_cumulatives

    def _window_diagonal_cumulatives(self, rotated_jisih):
        # An element of the first axis is defined as the sum of indices
        # for previous and subsequent ISI.

        jisih_diag_cums = np.zeros((self.n_bins,
                                    self.n_bins
                                    + 2 * self._max_change_index))

        # double_index corresponds to the sum of the indices for the previous
        # and the subsequent ISI.
        for double_index in range(self.n_bins):
            anti_diagonal = np.diagonal(
                rotated_jisih, - self.n_bins + double_index + 1)

            right_padding = jisih_diag_cums.shape[1] - \
                len(anti_diagonal) - self._max_change_index

            cumulated_diagonal = np.cumsum(anti_diagonal)

            padded_cumulated_diagonal = np.pad(
                cumulated_diagonal,
                pad_width=(self._max_change_index, right_padding),
                mode='constant',
                constant_values=(0., cumulated_diagonal[-1]))

            jisih_diag_cums[double_index] = padded_cumulated_diagonal

        return jisih_diag_cums

    def _get_dithered_isi(self, isi_to_dither):
        dithered_isi = np.copy(isi_to_dither)
        # if alternate is true, a sampling_rhythm of 2 means that we have two
        # partitions of spikes, the "odd" and the "even" spikes and first
        # dither the "even" ones and then the "odd" ones.
        # if alternate is false, we just go dither from the first to the last
        # spike, which corresponds to a sampling_rhythm of 1.
        sampling_rhythm = self.alternate + 1
        number_of_isis = len(dithered_isi)

        for start in range(sampling_rhythm):
            dithered_isi_indices = self._isi_to_index(dithered_isi)
            for i in range(start, number_of_isis - 1,
                           sampling_rhythm):
                step = self._get_dithering_step(
                    dithered_isi,
                    dithered_isi_indices,
                    i)
                dithered_isi[i] += step
                dithered_isi[i + 1] -= step

        return dithered_isi

    def _get_dithering_step(self,
                            dithered_isi,
                            dithered_isi_indices,
                            i):
        curr_isi_id = dithered_isi_indices[i]
        next_isi_id = dithered_isi_indices[i + 1]
        double_index = curr_isi_id + next_isi_id
        if double_index < self.n_bins:
            if self.method == 'fast':
                cum_dist_func = self._jisih_cumulatives[
                    double_index]
                compare_isi = self._index_to_isi(curr_isi_id + 1)
            else:
                cum_dist_func = self._jisih_cumulatives[
                    curr_isi_id][next_isi_id]
                compare_isi = self._max_change_isi

            if cum_dist_func[-1] > 0.:
                # when the method is 'fast', new_isi_id is where the current
                # ISI id should go to.
                new_isi_id = np.searchsorted(cum_dist_func, random.random())
                step = self._index_to_isi(new_isi_id)\
                    - compare_isi
                return step

        return self._uniform_dither_not_jisi_movable_spikes(
            dithered_isi[i],
            dithered_isi[i + 1])

    def _uniform_dither_not_jisi_movable_spikes(self,
                                                curr_isi,
                                                next_isi):
        left_dither = min(curr_isi - self.refractory_period, self.dither)
        right_dither = min(next_isi - self.refractory_period, self.dither)
        step = random.random() * (right_dither + left_dither) - left_dither
        return step


def trial_shifting(spiketrains, dither, n_surrogates=1):
    """
    Generates surrogates of a spike train by trial shifting.

    It shifts by a random uniform amount independently different trials,
    which are the elements of a list of spiketrains.

    The shifting is done independently for each surrogate.

    Parameters
    ----------
    spiketrains : list of neo.SpikeTrain
        A list of spike trains of the same neuron
        where each element corresponds to one trial.
    dither : pq.Quantity
        Amount of dithering.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1

    Returns
    -------
    surrogate_spiketrains : list of list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        randomly dithering its spikes. The range of the surrogate spike trains
        is the same as of `spiketrain`.
    """
    dither = dither.simplified.magnitude

    units = spiketrains[0].units
    t_starts = [single_trial_st.t_start.simplified.magnitude
                for single_trial_st in spiketrains]
    t_stops = [single_trial_st.t_stop.simplified.magnitude
               for single_trial_st in spiketrains]
    sampling_rates = [single_trial_st.sampling_rate
                      for single_trial_st in spiketrains]
    spiketrains = [single_trial_st.simplified.magnitude
                   for single_trial_st in spiketrains]

    surrogate_spiketrains = \
        _trial_shifting(spiketrains, dither, t_starts, t_stops,
                        n_surrogates)

    surrogate_spiketrains = \
        [[neo.SpikeTrain(
            surrogate_spiketrain[trial_id] * pq.s,
            t_start=t_starts[trial_id] * pq.s,
            t_stop=t_stops[trial_id] * pq.s,
            units=units,
            sampling_rate=sampling_rates[trial_id])
          for trial_id in range(len(surrogate_spiketrain))]
         for surrogate_spiketrain in surrogate_spiketrains]

    return surrogate_spiketrains


def _trial_shifting(spiketrains, dither, t_starts, t_stops, n_surrogates):
    """
    Inner nucleus of the trial shuffling procedure.
    """
    surrogate_spiketrains = []
    for surrogate_id in range(n_surrogates):
        copied_spiketrain = copy.deepcopy(spiketrains)
        surrogate_spiketrain = []
        # looping over all trials
        for trial_id, single_trial_st in enumerate(copied_spiketrain):
            single_trial_st += dither * (2 * random.random() - 1)
            single_trial_st = np.remainder(
                single_trial_st - t_starts[trial_id],
                t_stops[trial_id] - t_starts[trial_id]
            ) + t_starts[trial_id]
            single_trial_st.sort()

            surrogate_spiketrain.append(single_trial_st)

        surrogate_spiketrains.append(surrogate_spiketrain)
    return surrogate_spiketrains


def _trial_shifting_of_concatenated_spiketrain(
        spiketrain, dither, trial_length, trial_separation, n_surrogates=1):
    """
    Generates surrogates of a spike train by trial shifting.

    It shifts by a random uniform amount independently different trials,
    individuated by the `trial_length` and the possible buffering period
    `trial_separation` present in between trials.

    The shifting is done independently for each surrogate.

    Parameters
    ----------
    spiketrain : neo.SpikeTrain
        A single spike train, where the trials are concatenated.
    dither : pq.Quantity
        Amount of dithering.
    trial_length : pq.Quantity
        The length of the single-trial spiketrain.
    trial_separation : pq.Quantity
        Buffering in between trials in the concatenation of the spiketrain.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1

    Returns
    -------
    surrogate_spiketrains : list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain` by
        randomly dithering its spikes. The range of the surrogate spike trains
        is the same as of `spiketrain`.
    """
    units = spiketrain.units
    t_start = spiketrain.t_start.simplified.magnitude
    t_stop = spiketrain.t_stop.simplified.magnitude
    trial_length = trial_length.simplified.magnitude
    trial_separation = trial_separation.simplified.magnitude
    dither = dither.simplified.magnitude
    n_trials = int((t_stop - t_start) // (trial_length + trial_separation))
    t_starts = t_start + \
        np.arange(n_trials) * (trial_length + trial_separation)
    t_stops = t_starts + trial_length
    spiketrains = spiketrain.simplified.magnitude
    spiketrains = [spiketrains[(spiketrains >= t_starts[trial_id]) &
                               (spiketrains <= t_stops[trial_id])]
                   for trial_id in range(n_trials)]

    surrogate_spiketrains = _trial_shifting(
        spiketrains, dither, t_starts, t_stops, n_surrogates)

    surrogate_spiketrains = [neo.SpikeTrain(
        np.hstack(surrogate_spiketrain) * pq.s,
        t_start=t_start * pq.s,
        t_stop=t_stop * pq.s,
        units=units,
        sampling_rate=spiketrain.sampling_rate)
        for surrogate_spiketrain in surrogate_spiketrains]
    return surrogate_spiketrains


@deprecated_alias(n='n_surrogates', surr_method='method')
def surrogates(spiketrain, n_surrogates=1, method='dither_spike_train',
               dt=None, **kwargs):
    """
    Generates surrogates of a `spiketrain` by a desired generation
    method.

    This routine is a wrapper for the other surrogate generators in the
    module.

    The surrogates retain the `spiketrain.t_start` and `spiketrain.t_stop` of
    the original `spiketrain`.


    Parameters
    ----------
    spiketrain : neo.SpikeTrain or list of neo.SpikeTrain
        The spike train from which to generate the surrogates.
        The only method that accepts a list of spike trains instead of a single
        spike train to generate the surrogates from is 'trial_shifting'.
    n_surrogates : int, optional
        Number of surrogates to be generated.
        Default: 1
    method : str, optional
        The method to use to generate surrogate spike trains. Can be one of:

        * 'dither_spike_train': see `surrogates.dither_spike_train()`
            [`dt` needed]
        * 'dither_spikes': see `surrogates.dither_spikes()` [`dt` needed]
        * 'jitter_spikes': see `surrogates.jitter_spikes()` [`dt` needed]
        * 'randomise_spikes': see `surrogates.randomise_spikes()`
        * 'shuffle_isis': see `surrogates.shuffle_isis()`
        * 'joint_isi_dithering': see `surrogates.joint_isi_dithering()`
            [`dt` needed]
        * 'trial_shifting': see `surrogates.trial_shifting` [`dt` needed]
            If used on a neo.SpikeTrain, specify the key-word argument
            `trial_length` and `trial_separation` of type pq.Quantity.
            Else, `spiketrain` has to be a list of neo.SpikeTrain.
        * 'bin_shuffling': see `surrogates.bin_shuffling()` [`dt` needed]
            If used in this module, specify the key-word argument `bin_size`
            of type pq.Quantity.

        Default: 'dither_spike_train'
    dt : pq.Quantity, optional
        For methods shifting spike times or spike trains randomly around
        their original time
        (`dither_spikes`, `dither_spike_train`) or replacing them randomly
        within a certain window (`jitter_spikes`), dt represents the size of
        that shift / window. For other methods, dt is ignored.
        Default: None
    kwargs
        Keyword arguments passed to the chosen surrogate method.

    Returns
    -------
    list of neo.SpikeTrain
        Each surrogate spike train obtained independently from `spiketrain`
        according to chosen surrogate type. The time range of the surrogate
        spike trains is the same as in `spiketrain`.

    Raises
    ------
    TypeError
        If `spiketrain` is not either a `neo.SpikeTrain` object or a list of
        `neo.SpikeTrain`.

    ValueError
        If `method` is not one of the surrogate methods defined in this module.

        If `dt` is None and `method` is not 'randomise_spikes' nor
        'shuffle_isis'.
    """

    if isinstance(spiketrain, list):
        if not isinstance(spiketrain[0], neo.SpikeTrain):
            raise TypeError('spiketrain must be an instance neo.SpikeTrain or'
                            ' a list of neo.SpikeTrain')
    elif not isinstance(spiketrain, neo.SpikeTrain):
        raise TypeError('spiketrain must be an instance neo.SpikeTrain or'
                        ' a list of neo.SpikeTrain')

    if method == "dither_spikes_with_refractory_period":
        warnings.warn("'dither_spikes_with_refractory_period' is deprecated "
                      "in favor of 'dither_spikes'", DeprecationWarning)

    # Define the surrogate function to use, depending on the specified method
    surrogate_types = {
        'dither_spike_train': dither_spike_train,
        'dither_spikes': dither_spikes,
        'dither_spikes_with_refractory_period': dither_spikes,
        'jitter_spikes': jitter_spikes,
        'randomise_spikes': randomise_spikes,
        'shuffle_isis': shuffle_isis,
        'bin_shuffling': bin_shuffling,
        'trial_shifting': trial_shifting,
        'joint_isi_dithering': lambda n: JointISI(
            spiketrain, **kwargs).dithering(n),
        'isi_dithering': lambda n: JointISI(
            spiketrain, isi_dithering=True, **kwargs).dithering(n)
    }

    if method not in surrogate_types.keys():
        raise ValueError("Specified surrogate method ('{}') "
                         "is not valid".format(method))
    method = surrogate_types[method]

    if dt is None and method not in (randomise_spikes, shuffle_isis):
        raise ValueError(f"'{method.__name__}' method requires 'dt' parameter "
                         f"to be set")

    if method in (dither_spike_train, dither_spikes):
        return method(
            spiketrain, dt, n_surrogates=n_surrogates, **kwargs)
    if method in (randomise_spikes, shuffle_isis):
        return method(spiketrain, n_surrogates=n_surrogates, **kwargs)
    if method is jitter_spikes:
        return method(spiketrain, dt, n_surrogates=n_surrogates)
    if method is trial_shifting:
        if isinstance(spiketrain, list):
            return method(
                spiketrain, dither=dt, n_surrogates=n_surrogates)
        return _trial_shifting_of_concatenated_spiketrain(
            spiketrain, dither=dt, n_surrogates=n_surrogates, **kwargs)
    if method is bin_shuffling:
        max_displacement = int(
            dt.simplified.magnitude / kwargs['bin_size'].simplified.magnitude)
        return method(
            spiketrain, max_displacement=max_displacement,
            bin_size=kwargs['bin_size'], n_surrogates=n_surrogates)
    # surr_method is 'joint_isi_dithering' or isi_dithering:
    return method(n_surrogates)



# -*- coding: utf-8 -*-
"""
Unit tests for the phase analysis module.

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
from __future__ import division, print_function

import unittest

from neo import SpikeTrain, AnalogSignal
import numpy as np
import quantities as pq

import elephant.phase_analysis

from numpy.ma.testutils import assert_allclose


class SpikeTriggeredPhaseTestCase(unittest.TestCase):

    def setUp(self):
        tlen0 = 100 * pq.s
        f0 = 20. * pq.Hz
        fs0 = 1 * pq.ms
        t0 = np.arange(
            0, tlen0.rescale(pq.s).magnitude,
            fs0.rescale(pq.s).magnitude) * pq.s
        self.anasig0 = AnalogSignal(
            np.sin(2 * np.pi * (f0 * t0).simplified.magnitude),
            units=pq.mV, t_start=0 * pq.ms, sampling_period=fs0)
        self.st0 = SpikeTrain(
            np.arange(50, tlen0.rescale(pq.ms).magnitude - 50, 50) * pq.ms,
            t_start=0 * pq.ms, t_stop=tlen0)
        self.st1 = SpikeTrain(
            [100., 100.1, 100.2, 100.3, 100.9, 101.] * pq.ms,
            t_start=0 * pq.ms, t_stop=tlen0)

    def test_perfect_locking_one_spiketrain_one_signal(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            self.st0,
            interpolate=True)

        assert_allclose(phases[0], - np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_perfect_locking_many_spiketrains_many_signals(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0, self.st0],
            interpolate=True)

        assert_allclose(phases[0], -np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_perfect_locking_one_spiketrains_many_signals(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0],
            interpolate=True)

        assert_allclose(phases[0], -np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_perfect_locking_many_spiketrains_one_signal(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            [self.st0, self.st0],
            interpolate=True)

        assert_allclose(phases[0], -np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_interpolate(self):
        phases_int, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            self.st1,
            interpolate=True)

        self.assertLess(phases_int[0][0], phases_int[0][1])
        self.assertLess(phases_int[0][1], phases_int[0][2])
        self.assertLess(phases_int[0][2], phases_int[0][3])
        self.assertLess(phases_int[0][3], phases_int[0][4])
        self.assertLess(phases_int[0][4], phases_int[0][5])

        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            self.st1,
            interpolate=False)

        self.assertEqual(phases_noint[0][0], phases_noint[0][1])
        self.assertEqual(phases_noint[0][1], phases_noint[0][2])
        self.assertEqual(phases_noint[0][2], phases_noint[0][3])
        self.assertEqual(phases_noint[0][3], phases_noint[0][4])
        self.assertNotEqual(phases_noint[0][4], phases_noint[0][5])

        # Verify that when using interpolation and the spike sits on the sample
        # of the Hilbert transform, this is the same result as when not using
        # interpolation with a spike slightly to the right
        self.assertEqual(phases_noint[0][2], phases_int[0][0])
        self.assertEqual(phases_noint[0][4], phases_int[0][0])

    def test_inconsistent_numbers_spiketrains_hilbert(self):
        self.assertRaises(
            ValueError, elephant.phase_analysis.spike_triggered_phase,
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0, self.st0, self.st0], False)

        self.assertRaises(
            ValueError, elephant.phase_analysis.spike_triggered_phase,
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0, self.st0, self.st0], False)

    def test_spike_earlier_than_hilbert(self):
        # This is a spike clearly outside the bounds
        st = SpikeTrain(
            [-50, 50],
            units='s', t_start=-100 * pq.s, t_stop=100 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 1)

        # This is a spike right on the border (start of the signal is at 0s,
        # spike sits at t=0s). By definition of intervals in
        # Elephant (left borders inclusive, right borders exclusive), this
        # spike is to be considered.
        st = SpikeTrain(
            [0, 50],
            units='s', t_start=-100 * pq.s, t_stop=100 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 2)

    def test_spike_later_than_hilbert(self):
        # This is a spike clearly outside the bounds
        st = SpikeTrain(
            [1, 250],
            units='s', t_start=-1 * pq.s, t_stop=300 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 1)

        # This is a spike right on the border (length of the signal is 100s,
        # spike sits at t=100s). However, by definition of intervals in
        # Elephant (left borders inclusive, right borders exclusive), this
        # spike is not to be considered.
        st = SpikeTrain(
            [1, 100],
            units='s', t_start=-1 * pq.s, t_stop=200 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 1)

    # This test handles the correct dealing with input signals that have
    # different time units, including a CompoundUnit
    def test_regression_269(self):
        # This is a spike train on a 30KHz sampling, one spike at 1s, one just
        # before the end of the signal
        cu = pq.CompoundUnit("1/30000.*s")
        st = SpikeTrain(
            [30000., (self.anasig0.t_stop - 1 * pq.s).rescale(cu).magnitude],
            units=pq.CompoundUnit("1/30000.*s"),
            t_start=-1 * pq.s, t_stop=300 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 2)


class MeanVectorTestCase(unittest.TestCase):
    def setUp(self):
        self.tolerance = 1e-15
        self.n_samples = 200
        # create a sample with all values equal to a random phase-lock phi
        self.lock_value_phi = np.random.uniform(-np.pi, np.pi, 1)
        self.dataset1 = np.ones(self.n_samples) * self.lock_value_phi
        # create a evenly spaced / uniform distribution
        self.dataset2 = np.arange(0, 2 * np.pi, (2 * np.pi) / self.n_samples)
        # create a random distribution
        self.dataset3 = np.random.uniform(-np.pi, np.pi, self.n_samples)

    def testMeanVector_direction_is_phi_and_length_is_1(self):
        """
        Test if the mean vector length is 1 and if the mean direction is phi
        for a sample with all phases equal to phi on the unit circle.

        """
        theta_bar_1, r_1 = elephant.phase_analysis.mean_phase_vector(
            self.dataset1)
        # mean direction must be phi
        self.assertAlmostEqual(theta_bar_1, self.lock_value_phi,
                               delta=self.tolerance)
        # mean vector length must be almost equal 1
        self.assertAlmostEqual(r_1, 1, delta=self.tolerance)

    def testMeanVector_length_is_0(self):
        """
        Test if the mean vector length  is 0 for a evenly spaced distribution
        on the unit circle.
        """
        theta_bar_2, r_2 = elephant.phase_analysis.mean_phase_vector(
            self.dataset2)
        # mean vector length must be almost equal 0
        self.assertAlmostEqual(r_2, 0, delta=self.tolerance)

    def testMeanVector_ranges_of_direction_and_length(self):
        """
        Test if the range of the mean vector direction follows numpy standard
        and is within (-pi, pi].
        Test if the range of the mean vector length is within [0, 1].
        """
        theta_bar_3, r_3 = elephant.phase_analysis.mean_phase_vector(
            self.dataset3)
        # mean vector direction
        self.assertTrue(-np.pi < theta_bar_3 <= np.pi)
        # mean vector length
        self.assertTrue(0 <= r_3 <= 1)


class PhaseDifferenceTestCase(unittest.TestCase):
    def setUp(self):
        self.tolerance = 1e-15
        self.n_samples = 200

    def testPhaseDifference_in_range_minus_pi_to_pi(self):
        """
        Test if the range of the phase difference is within [-pi, pi] for
        random pairs of alpha and beta.
        """
        alpha = np.random.uniform(-np.pi, np.pi, self.n_samples)
        beta = np.random.uniform(-np.pi, np.pi, self.n_samples)

        phase_diff = elephant.phase_analysis.phase_difference(alpha, beta)
        self.assertTrue((-np.pi <= phase_diff).all()
                        and (phase_diff <= np.pi).all())

    def testPhaseDifference_is_delta(self):
        """
        Test if the phase difference is random delta for random pairs of
        alpha and beta, where beta is a copy of alpha shifted by delta.
        """
        delta = np.random.uniform(-np.pi, np.pi, self.n_samples)
        alpha = np.random.uniform(-np.pi, np.pi, self.n_samples)
        _beta = alpha - delta
        beta = np.arctan2(np.sin(_beta), np.cos(_beta))

        phase_diff = elephant.phase_analysis.phase_difference(alpha, beta)
        np.testing.assert_allclose(phase_diff, delta, atol=1e-10)


class PhaseLockingValueTestCase(unittest.TestCase):
    def setUp(self):
        self.tolerance = 1e-15
        self.phase_shift = np.pi / 4
        self.num_time_points = 1000
        self.num_trials = 100

        # create two random uniform distributions (all trials are identical)
        self.signal_x = \
            np.full([self.num_trials, self.num_time_points],
                    np.random.uniform(-np.pi, np.pi, self.num_time_points))
        self.signal_y = \
            np.full([self.num_trials, self.num_time_points],
                    np.random.uniform(-np.pi, np.pi, self.num_time_points))

        # create two random uniform distributions, where all trails are random
        self.random_x = np.random.uniform(
            -np.pi, np.pi, (1000, self.num_time_points))
        self.random_y = np.random.uniform(
            -np.pi, np.pi, (1000, self.num_time_points))

        # simple samples of different shapes to assert ErrorRaising
        self.simple_x = np.array([[0, -np.pi, np.pi], [0, -np.pi, np.pi]])
        self.simple_y = np.array([0, -np.pi, np.pi])
        self.simple_z = np.array([0, np.pi, np.pi / 2, -np.pi])

    def testPhaseLockingValue_identical_signals_both_identical_trials(self):
        """
        Test if the PLV's are 1, when 2 identical signals with identical
        trials are passed. PLV's needed to be 1, due to the constant phase
        difference of 0 across trials at each time-point.
        """
        list1_plv_t = \
            elephant.phase_analysis.phase_locking_value(self.signal_x,
                                                        self.signal_x)
        target_plv_r_is_one = np.ones_like(list1_plv_t)
        np.testing.assert_allclose(list1_plv_t, target_plv_r_is_one,
                                   self.tolerance)

    def testPhaseLockingValue_different_signals_both_identical_trials(self):
        """
        Test if the PLV's are 1, when 2 different signals are passed, where
        within each signal the trials are identical. PLV's needed to be 1,
        due to a constant phase difference across trials, which may vary for
        different time-points.
        """
        list2_plv_t = elephant.phase_analysis.phase_locking_value(
            self.signal_x, self.signal_y)
        target_plv_r_is_one = np.ones_like(list2_plv_t)
        np.testing.assert_allclose(list2_plv_t, target_plv_r_is_one,
                                   atol=3e-15)

    def testPhaseLockingValue_different_signals_both_different_trials(self):
        """
        Test if the PLV's are close to 0, when 2 different signals are passed,
        where both have different trials, which are all randomly distributed.
        The PLV's needed to be close to 0, do to a random
        phase difference across trials for each time-point.
        """
        list3_plv_t = elephant.phase_analysis.phase_locking_value(
            self.random_x, self.random_y)
        target_plv_is_zero = np.zeros_like(list3_plv_t)
        # use default value from np.allclose() for atol=1e-8 to prevent failure
        np.testing.assert_allclose(list3_plv_t, target_plv_is_zero,
                                   rtol=1e-2, atol=1.1e-1)

    def testPhaseLockingValue_raise_Error_if_trial_number_is_different(self):
        """
        Test if a ValueError is raised, when the signals have different
        number of trails.
        """
        # different numbers of trails
        np.testing.assert_raises(
            ValueError, elephant.phase_analysis.phase_locking_value,
            self.simple_x, self.simple_y)

    def testPhaseLockingValue_raise_Error_if_trial_lengths_are_different(self):
        """
        Test if a ValueError is raised, when within a trail-pair of the signals
        the trial-lengths are different.
        """
        # different lengths in a trail pair
        np.testing.assert_raises(
            ValueError, elephant.phase_analysis.phase_locking_value,
            self.simple_y, self.simple_z)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
This script is used to generate basis sources for the
kCSD method Jan et.al (2012) for 1D,2D and 3D cases.
Two 'types' are described here, gaussian and step source,
These can be easily extended.
These scripts are based on Grzegorz Parka's,
Google Summer of Code 2014, INFC/pykCSD
This was written by :
Michal Czerwinski, Chaitanya Chintaluri
Laboratory of Neuroinformatics,
Nencki Institute of Experimental Biology, Warsaw.
"""
from __future__ import division

import numpy as np

def gauss(d, stdev, dim):
    """Gaussian function
    Parameters
    ----------
    d : floats or np.arrays
        Distance array to the point of evaluation
    stdev : float
        cutoff range
    dim : int
        dimension of the gaussian function
    Returns
    -------
    Z : floats or np.arrays
        function evaluated
    """
    Z = np.exp(-(d**2) / (2* stdev**2) ) / (np.sqrt(2*np.pi)*stdev)**dim
    return Z

def step_1D(d, R):
    """Returns normalized 1D step function.
    Parameters
    ----------
    d : floats or np.arrays
        Distance array to the point of evaluation
    R : float
        cutoff range
    Returns
    -------
    s : Value of the function (d  <= R) / R
    """
    s = (d  <= R)
    s = s / R #normalize with width
    return s

def gauss_1D(d, three_stdev):
    """Returns normalized gaussian 2D scale function
    Parameters
    ----------
    d : floats or np.arrays
        Distance array to the point of evaluation
    three_stdev : float
        3 * standard deviation of the distribution
    Returns
    -------
    Z : (three_std/3)*(1/2*pi)*(exp(-0.5)*stddev**(-2) *(d**2))
    """
    stdev = three_stdev/3.0
    Z = gauss(d, stdev, 1)
    return Z

def gauss_lim_1D(d, three_stdev):
    """Returns gausian 2D function cut off after 3 standard deviations.
    Parameters
    ----------
    d : floats or np.arrays
        Distance array to the point of evaluation
    three_stdev : float
        3 * standard deviation of the distribution
    Returns
    -------
    Z : (three_std/3)*(1/2*pi)*(exp(-0.5)*stddev**(-2) *((x-mu)**2)),
        cut off = three_stdev
    """
    Z = gauss_1D(d, three_stdev)
    Z *= (d < three_stdev)
    return Z

def step_2D(d, R):
    """Returns normalized 2D step function.
    Parameters
    ----------
    d : float or np.arrays
        Distance array to the point of evaluation
    R : float
        cutoff range

    Returns
    -------
    s : step function
    """
    s = (d <= R) / (np.pi*(R**2))
    return s

def gauss_2D(d, three_stdev):
    """Returns normalized gaussian 2D scale function
    Parameters
    ----------
    d : floats or np.arrays
         distance at which we need the function evaluated
    three_stdev : float
        3 * standard deviation of the distribution
    Returns
    -------
    Z : function
        Normalized gaussian 2D function
    """
    stdev = three_stdev/3.0
    Z = gauss(d, stdev, 2)
    return Z

def gauss_lim_2D(d, three_stdev):
    """Returns gausian 2D function cut off after 3 standard deviations.
    Parameters
    ----------
    d : floats or np.arrays
         distance at which we need the function evaluated
    three_stdev : float
        3 * standard deviation of the distribution
    Returns
    -------
    Z : function
        Normalized gaussian 2D function cut off after three_stdev
    """
    Z = (d <= three_stdev)*gauss_2D(d, three_stdev)
    return Z

def gauss_3D(d, three_stdev):
    """Returns normalized gaussian 3D scale function
    Parameters
    ----------
    d : floats or np.arrays
        distance at which we need the function evaluated
    three_stdev : float
        3 * standard deviation of the distribution
    Returns
    -------
    Z : funtion
        Normalized gaussian 3D function
    """
    stdev = three_stdev/3.0
    Z = gauss(d, stdev, 3)
    return Z

def gauss_lim_3D(d, three_stdev):
    """Returns normalized gaussian 3D scale function cut off after 3stdev
    Parameters
    ----------
    d : floats or np.arrays
        distance at which we need the function evaluated
    three_stdev : float
        3 * standard deviation of the distribution
    Returns
    -------
    Z : funtion
        Normalized gaussian 3D function cutoff three_Stdev
    """
    Z = gauss_3D(d, three_stdev)
    Z = Z * (d < (three_stdev))
    return Z

def step_3D(d, R):
    """Returns normalized 3D step function.
    Parameters
    ----------
    d : floats or np.arrays
        distance at which we need the function evaluated
    R : float
        cutoff range
    Returns
    -------
    s : step function in 3D
    """

    s = 3/(4*np.pi*R**3)*(d <= R)
    return s

basis_1D = {
    "step": step_1D,
    "gauss": gauss_1D,
    "gauss_lim": gauss_lim_1D,
}


basis_2D = {
    "step": step_2D,
    "gauss": gauss_2D,
    "gauss_lim": gauss_lim_2D,
}

basis_3D = {
    "step": step_3D,
    "gauss": gauss_3D,
    "gauss_lim": gauss_lim_3D,
}



# -*- coding: utf-8 -*-
"""
Unit tests for the spectral module.

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

import unittest

import numpy as np
import scipy.signal as spsig
import quantities as pq
import neo.core as n
from numpy.testing import assert_array_almost_equal, assert_array_equal

import elephant.spectral
from elephant.datasets import download_datasets, ELEPHANT_TMP_DIR


class WelchPSDTestCase(unittest.TestCase):
    def test_welch_psd_errors(self):
        # generate a dummy data
        data = n.AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                              units='mV')

        # check for invalid parameter values
        # - length of segments
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          len_seg=0)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          len_seg=data.shape[0] * 2)
        # - number of segments
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          num_seg=0)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          num_seg=data.shape[0] * 2)
        # - frequency resolution
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          freq_res=-1)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          freq_res=data.sampling_rate / (data.shape[0] + 1))
        # - overlap
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          overlap=-1.0)
        self.assertRaises(ValueError, elephant.spectral.welch_psd, data,
                          overlap=1.1)

    def test_welch_psd_behavior(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=data_length)
        signal = [np.sin(2 * np.pi * signal_freq * t)
                  for t in np.arange(0, data_length * sampling_period,
                                     sampling_period)]
        data = n.AnalogSignal(np.array(signal + noise),
                              sampling_period=sampling_period * pq.s,
                              units='mV')

        # consistency between different ways of specifying segment length
        freqs1, psd1 = elephant.spectral.welch_psd(
            data, len_segment=data_length // 5, overlap=0)
        freqs2, psd2 = elephant.spectral.welch_psd(
            data, n_segments=5, overlap=0)
        self.assertTrue((psd1 == psd2).all() and (freqs1 == freqs2).all())

        # frequency resolution and consistency with data
        freq_res = 1.0 * pq.Hz
        freqs, psd = elephant.spectral.welch_psd(
            data, frequency_resolution=freq_res)
        self.assertAlmostEqual(freq_res, freqs[1] - freqs[0])
        self.assertEqual(freqs[psd.argmax()], signal_freq)
        freqs_np, psd_np = elephant.spectral.welch_psd(
            data.magnitude.flatten(), fs=1 / sampling_period,
            frequency_resolution=freq_res)
        self.assertTrue((freqs == freqs_np).all() and (psd == psd_np).all())

        # check of scipy.signal.welch() parameters
        params = {'window': 'hamming', 'nfft': 1024, 'detrend': 'linear',
                  'return_onesided': False, 'scaling': 'spectrum'}
        for key, val in params.items():
            freqs, psd = elephant.spectral.welch_psd(
                data, len_segment=1000, overlap=0, **{key: val})
            freqs_spsig, psd_spsig = spsig.welch(np.rollaxis(data, 0, len(
                data.shape)), fs=1 / sampling_period, nperseg=1000,
                                                 noverlap=0, **{key: val})
            self.assertTrue(
                (freqs == freqs_spsig).all() and (
                        psd == psd_spsig).all())

        # - generate multidimensional data for check of parameter `axis`
        num_channel = 4
        data_length = 5000
        data_multidim = np.random.normal(size=(num_channel, data_length))
        freqs, psd = elephant.spectral.welch_psd(data_multidim)
        freqs_T, psd_T = elephant.spectral.welch_psd(data_multidim.T, axis=0)
        self.assertTrue(np.all(freqs == freqs_T))
        self.assertTrue(np.all(psd == psd_T.T))

    def test_welch_psd_input_types(self):
        # generate a test data
        sampling_period = 0.001
        data = n.AnalogSignal(np.array(np.random.normal(size=5000)),
                              sampling_period=sampling_period * pq.s,
                              units='mV')

        # outputs from AnalogSignal input are of Quantity type (standard usage)
        freqs_neo, psd_neo = elephant.spectral.welch_psd(data)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, psd_pq = elephant.spectral.welch_psd(
            data.magnitude.flatten() * data.units, fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, psd_np = elephant.spectral.welch_psd(
            data.magnitude.flatten(), fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(psd_np, pq.quantity.Quantity))

        # check if the results from different input types are identical
        self.assertTrue(
            (freqs_neo == freqs_pq).all() and (
                    psd_neo == psd_pq).all())
        self.assertTrue(
            (freqs_neo == freqs_np).all() and (
                    psd_neo == psd_np).all())

    def test_welch_psd_multidim_input(self):
        # generate multidimensional data
        num_channel = 4
        data_length = 5000
        sampling_period = 0.001
        noise = np.random.normal(size=(num_channel, data_length))
        data_np = np.array(noise)
        # Since row-column order in AnalogSignal is different from the
        # conventional one, `data_np` needs to be transposed when its used to
        # define an AnalogSignal
        data_neo = n.AnalogSignal(data_np.T,
                                  sampling_period=sampling_period * pq.s,
                                  units='mV')
        data_neo_1dim = n.AnalogSignal(data_np[0],
                                       sampling_period=sampling_period * pq.s,
                                       units='mV')

        # check if the results from different input types are identical
        freqs_np, psd_np = elephant.spectral.welch_psd(data_np,
                                                       fs=1 / sampling_period)
        freqs_neo, psd_neo = elephant.spectral.welch_psd(data_neo)
        freqs_neo_1dim, psd_neo_1dim = elephant.spectral.welch_psd(
            data_neo_1dim)
        self.assertTrue(np.all(freqs_np == freqs_neo))
        self.assertTrue(np.all(psd_np == psd_neo))
        self.assertTrue(np.all(psd_neo_1dim == psd_neo[0]))


class MultitaperPSDTestCase(unittest.TestCase):
    def test_multitaper_psd_errors(self):
        # generate dummy data
        signal = n.AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                                units='mV')
        fs = 1000 * pq.Hz
        nw = 3

        # check for invalid parameter values
        # - number of tapers
        self.assertRaises(ValueError, elephant.spectral.multitaper_psd, signal,
                          fs, nw, num_tapers=-5)
        self.assertRaises(TypeError, elephant.spectral.multitaper_psd, signal,
                          fs, nw, num_tapers=-5.0)
        # - frequency resolution
        self.assertRaises(ValueError, elephant.spectral.multitaper_psd, signal,
                          fs, nw, peak_resolution=-1)

    def test_multitaper_psd_behavior(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=data_length)
        signal = [np.sin(2 * np.pi * signal_freq * t)
                  for t in np.arange(0, data_length * sampling_period,
                                     sampling_period)]
        data = n.AnalogSignal(np.array(signal + noise),
                              sampling_period=sampling_period * pq.s,
                              units='mV')

        # consistency between different ways of specifying number of tapers
        freqs1, psd1 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3.5)
        freqs2, psd2 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3.5,
                                                        num_tapers=6)
        self.assertTrue((psd1 == psd2).all() and (freqs1 == freqs2).all())

        # frequency resolution and consistency with data
        freq_res = 1.0 * pq.Hz
        freqs, psd = elephant.spectral.multitaper_psd(
            data, peak_resolution=freq_res)
        self.assertEqual(freqs[psd.argmax()], signal_freq)
        freqs_np, psd_np = elephant.spectral.multitaper_psd(
            data.magnitude.flatten(), fs=1 / sampling_period,
            peak_resolution=freq_res)
        self.assertTrue((freqs == freqs_np).all() and (psd == psd_np).all())

    def test_multitaper_psd_parameter_hierarchy(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise = np.random.normal(size=data_length)
        signal = [np.sin(2 * np.pi * signal_freq * t)
                  for t in np.arange(0, data_length * sampling_period,
                                     sampling_period)]
        data = n.AnalogSignal(np.array(signal + noise),
                              sampling_period=sampling_period * pq.s,
                              units='mV')

        # Test num_tapers vs nw
        freqs1, psd1 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3,
                                                        num_tapers=9)
        freqs2, psd2 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3)
        self.assertTrue((freqs1 == freqs2).all() and (psd1 != psd2).all())

        # Test peak_resolution vs nw
        freqs1, psd1 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3,
                                                        num_tapers=9,
                                                        peak_resolution=1)
        freqs2, psd2 = elephant.spectral.multitaper_psd(data,
                                                        fs=data.sampling_rate,
                                                        nw=3,
                                                        num_tapers=9)
        self.assertTrue((freqs1 == freqs2).all() and (psd1 != psd2).all())

    def test_multitaper_psd_against_nitime(self):
        """
        This test assesses the match between this implementation of
        multitaper against nitime (0.8) using a predefined time series
        generated by an autoregressive model.

        Please follow the link below for more details:
        https://gin.g-node.org/INM-6/elephant-data/src/master/unittest/spectral/multitaper_psd
        """
        repo_path = r"unittest/spectral/multitaper_psd/data"

        files_to_download = [
            ("time_series.npy", "ff43797e2ac94613f510b20a31e2e80e"),
            ("psd_nitime.npy", "89d1f53957e66c786049ea425b53c0e8")
        ]

        for filename, checksum in files_to_download:
            download_datasets(repo_path=f"{repo_path}/{filename}",
                              checksum=checksum)

        time_series = np.load(ELEPHANT_TMP_DIR / 'time_series.npy')
        psd_nitime = np.load(ELEPHANT_TMP_DIR / 'psd_nitime.npy')

        freqs, psd_multitaper = elephant.spectral.multitaper_psd(
            signal=time_series, fs=0.1, nw=4, num_tapers=8)

        np.testing.assert_allclose(psd_multitaper, psd_nitime, rtol=0.3,
                                   atol=0.1)

    def test_multitaper_psd_input_types(self):
        # generate a test data
        sampling_period = 0.001
        data = n.AnalogSignal(np.array(np.random.normal(size=5000)),
                              sampling_period=sampling_period * pq.s,
                              units='mV')

        # outputs from AnalogSignal input are of Quantity type (standard usage)
        freqs_neo, psd_neo = elephant.spectral.multitaper_psd(data)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, psd_pq = elephant.spectral.multitaper_psd(
            data.magnitude.flatten() * data.units, fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(psd_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, psd_np = elephant.spectral.multitaper_psd(
            data.magnitude.flatten(), fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(psd_np, pq.quantity.Quantity))

        # check if the results from different input types are identical
        self.assertTrue(
            (freqs_neo == freqs_pq).all() and (
                    psd_neo == psd_pq).all())
        self.assertTrue(
            (freqs_neo == freqs_np).all() and (
                    psd_neo == psd_np).all())


class WelchCohereTestCase(unittest.TestCase):
    def test_welch_cohere_errors(self):
        # generate a dummy data
        x = n.AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                           units='mV')
        y = n.AnalogSignal(np.zeros(5000), sampling_period=0.001 * pq.s,
                           units='mV')

        # check for invalid parameter values
        # - length of segments
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          len_seg=0)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          len_seg=x.shape[0] * 2)
        # - number of segments
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          num_seg=0)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          num_seg=x.shape[0] * 2)
        # - frequency resolution
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          freq_res=-1)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          freq_res=x.sampling_rate / (x.shape[0] + 1))
        # - overlap
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          overlap=-1.0)
        self.assertRaises(ValueError, elephant.spectral.welch_coherence, x, y,
                          overlap=1.1)

    def test_welch_cohere_behavior(self):
        # generate data by adding white noise and a sinusoid
        data_length = 5000
        sampling_period = 0.001
        signal_freq = 100.0
        noise1 = np.random.normal(size=data_length) * 0.01
        noise2 = np.random.normal(size=data_length) * 0.01
        signal1 = [np.cos(2 * np.pi * signal_freq * t)
                   for t in np.arange(0, data_length * sampling_period,
                                      sampling_period)]
        signal2 = [np.sin(2 * np.pi * signal_freq * t)
                   for t in np.arange(0, data_length * sampling_period,
                                      sampling_period)]
        x = n.AnalogSignal(np.array(signal1 + noise1), units='mV',
                           sampling_period=sampling_period * pq.s)
        y = n.AnalogSignal(np.array(signal2 + noise2), units='mV',
                           sampling_period=sampling_period * pq.s)

        # consistency between different ways of specifying segment length
        freqs1, coherency1, phase_lag1 = elephant.spectral.welch_coherence(
            x, y, len_segment=data_length // 5, overlap=0)
        freqs2, coherency2, phase_lag2 = elephant.spectral.welch_coherence(
            x, y, n_segments=5, overlap=0)
        self.assertTrue((coherency1 == coherency2).all() and
                        (phase_lag1 == phase_lag2).all() and
                        (freqs1 == freqs2).all())

        # frequency resolution and consistency with data
        freq_res = 1.0 * pq.Hz
        freqs, coherency, phase_lag = elephant.spectral.welch_coherence(
            x, y, frequency_resolution=freq_res)
        self.assertAlmostEqual(freq_res, freqs[1] - freqs[0])
        self.assertAlmostEqual(freqs[coherency.argmax()], signal_freq,
                               places=2)
        self.assertAlmostEqual(phase_lag[coherency.argmax()], -np.pi / 2,
                               places=2)
        freqs_np, coherency_np, phase_lag_np = \
            elephant.spectral.welch_coherence(x.magnitude.flatten(),
                                              y.magnitude.flatten(),
                                              fs=1 / sampling_period,
                                              frequency_resolution=freq_res)
        assert_array_equal(freqs.simplified.magnitude, freqs_np)
        assert_array_equal(coherency[:, 0], coherency_np)
        assert_array_equal(phase_lag[:, 0], phase_lag_np)

        # - check the behavior of parameter `axis` using multidimensional data
        num_channel = 4
        data_length = 5000
        x_multidim = np.random.normal(size=(num_channel, data_length))
        y_multidim = np.random.normal(size=(num_channel, data_length))
        freqs, coherency, phase_lag = \
            elephant.spectral.welch_coherence(x_multidim, y_multidim)
        freqs_T, coherency_T, phase_lag_T = elephant.spectral.welch_coherence(
            x_multidim.T, y_multidim.T, axis=0)
        assert_array_equal(freqs, freqs_T)
        assert_array_equal(coherency, coherency_T.T)
        assert_array_equal(phase_lag, phase_lag_T.T)

    def test_welch_cohere_input_types(self):
        # generate a test data
        sampling_period = 0.001
        x = n.AnalogSignal(np.array(np.random.normal(size=5000)),
                           sampling_period=sampling_period * pq.s,
                           units='mV')
        y = n.AnalogSignal(np.array(np.random.normal(size=5000)),
                           sampling_period=sampling_period * pq.s,
                           units='mV')

        # outputs from AnalogSignal input are of Quantity type
        # (standard usage)
        freqs_neo, coherency_neo, phase_lag_neo = \
            elephant.spectral.welch_coherence(x, y)
        self.assertTrue(isinstance(freqs_neo, pq.quantity.Quantity))
        self.assertTrue(isinstance(phase_lag_neo, pq.quantity.Quantity))

        # outputs from Quantity array input are of Quantity type
        freqs_pq, coherency_pq, phase_lag_pq = elephant.spectral \
            .welch_coherence(x.magnitude.flatten() * x.units,
                             y.magnitude.flatten() * y.units,
                             fs=1 / sampling_period)
        self.assertTrue(isinstance(freqs_pq, pq.quantity.Quantity))
        self.assertTrue(isinstance(phase_lag_pq, pq.quantity.Quantity))

        # outputs from Numpy ndarray input are NOT of Quantity type
        freqs_np, coherency_np, phase_lag_np = elephant.spectral \
            .welch_coherence(x.magnitude.flatten(),
                             y.magnitude.flatten(),
                             fs=1 / sampling_period)
        self.assertFalse(isinstance(freqs_np, pq.quantity.Quantity))
        self.assertFalse(isinstance(phase_lag_np, pq.quantity.Quantity))

        # check if the results from different input types are identical
        self.assertTrue((freqs_neo == freqs_pq).all() and
                        (coherency_neo[:, 0] == coherency_pq).all() and
                        (phase_lag_neo[:, 0] == phase_lag_pq).all())
        self.assertTrue((freqs_neo == freqs_np).all() and
                        (coherency_neo[:, 0] == coherency_np).all() and
                        (phase_lag_neo[:, 0] == phase_lag_np).all())

    def test_welch_cohere_multidim_input(self):
        # generate multidimensional data
        num_channel = 4
        data_length = 5000
        sampling_period = 0.001
        x_np = np.array(np.random.normal(size=(num_channel, data_length)))
        y_np = np.array(np.random.normal(size=(num_channel, data_length)))
        # Since row-column order in AnalogSignal is different from the
        # convention in NumPy/SciPy, `data_np` needs to be transposed when its
        # used to define an AnalogSignal
        x_neo = n.AnalogSignal(x_np.T, units='mV',
                               sampling_period=sampling_period * pq.s)
        y_neo = n.AnalogSignal(y_np.T, units='mV',
                               sampling_period=sampling_period * pq.s)
        x_neo_1dim = n.AnalogSignal(x_np[0], units='mV',
                                    sampling_period=sampling_period * pq.s)
        y_neo_1dim = n.AnalogSignal(y_np[0], units='mV',
                                    sampling_period=sampling_period * pq.s)

        # check if the results from different input types are identical
        freqs_np, coherency_np, phase_lag_np = elephant.spectral \
            .welch_coherence(x_np, y_np, fs=1 / sampling_period)
        freqs_neo, coherency_neo, phase_lag_neo = \
            elephant.spectral.welch_coherence(x_neo, y_neo)
        freqs_neo_1dim, coherency_neo_1dim, phase_lag_neo_1dim = \
            elephant.spectral.welch_coherence(x_neo_1dim, y_neo_1dim)
        self.assertTrue(np.all(freqs_np == freqs_neo))
        self.assertTrue(np.all(coherency_np.T == coherency_neo))
        self.assertTrue(np.all(phase_lag_np.T == phase_lag_neo))
        self.assertTrue(
            np.all(coherency_neo_1dim[:, 0] == coherency_neo[:, 0]))
        self.assertTrue(
            np.all(phase_lag_neo_1dim[:, 0] == phase_lag_neo[:, 0]))


if __name__ == "__main__":
    unittest.main(verbosity=2)



# -*- coding: utf-8 -*-
import unittest

import neo
import numpy as np
import quantities as pq
from numpy.testing import assert_array_almost_equal, assert_allclose

import elephant.change_point_detection as mft


class FilterTestCase(unittest.TestCase):
    def setUp(self):
        self.test_array = [0.4, 0.5, 0.65, 0.7, 0.9, 1.15, 1.2, 1.9]
        # spks_ri = [0.9, 1.15, 1.2]
        # spk_le = [0.4, 0.5, 0.65, 0.7]
        mu_ri = (0.25 + 0.05) / 2
        mu_le = (0.1 + 0.15 + 0.05) / 3
        sigma_ri = ((0.25 - 0.15) ** 2 + (0.05 - 0.15) ** 2) / 2
        sigma_le = ((0.1 - 0.1) ** 2 + (0.15 - 0.1) ** 2 + (
            0.05 - 0.1) ** 2) / 3
        self.targ_t08_h025 = 0
        self.targ_t08_h05 = (3 - 4) / np.sqrt(
            (sigma_ri / mu_ri ** 3) * 0.5 + (sigma_le / mu_le ** 3) * 0.5)

    # Window Large #
    def test_filter_with_spiketrain_h05(self):
        st = neo.SpikeTrain(self.test_array, units='s', t_stop=2.0)
        target = self.targ_t08_h05
        res = mft._filter(0.8 * pq.s, 0.5 * pq.s, st)
        assert_array_almost_equal(res, target, decimal=9)
        self.assertRaises(ValueError, mft._filter, 0.8, 0.5 * pq.s, st)
        self.assertRaises(ValueError, mft._filter, 0.8 * pq.s, 0.5, st)
        self.assertRaises(ValueError, mft._filter, 0.8 * pq.s, 0.5 * pq.s,
                          self.test_array)

    # Window Small #
    def test_filter_with_spiketrain_h025(self):
        st = neo.SpikeTrain(self.test_array, units='s', t_stop=2.0)
        target = self.targ_t08_h025
        res = mft._filter(0.8 * pq.s, 0.25 * pq.s, st)
        assert_array_almost_equal(res, target, decimal=9)

    def test_filter_with_quantities_h025(self):
        st = pq.Quantity(self.test_array, units='s')
        target = self.targ_t08_h025
        res = mft._filter(0.8 * pq.s, 0.25 * pq.s, st)
        assert_array_almost_equal(res, target, decimal=9)

    def test_filter_with_plain_array_h025(self):
        st = self.test_array
        target = self.targ_t08_h025
        res = mft._filter(0.8 * pq.s, 0.25 * pq.s, st * pq.s)
        assert_array_almost_equal(res, target, decimal=9)

    def test_isi_with_quantities_h05(self):
        st = pq.Quantity(self.test_array, units='s')
        target = self.targ_t08_h05
        res = mft._filter(0.8 * pq.s, 0.5 * pq.s, st)
        assert_array_almost_equal(res, target, decimal=9)

    def test_isi_with_plain_array_h05(self):
        st = self.test_array
        target = self.targ_t08_h05
        res = mft._filter(0.8 * pq.s, 0.5 * pq.s, st * pq.s)
        assert_array_almost_equal(res, target, decimal=9)


class FilterProcessTestCase(unittest.TestCase):
    def setUp(self):
        self.test_array = [1.1, 1.2, 1.4, 1.6, 1.7, 1.75, 1.8, 1.85, 1.9, 1.95]
        x = (7 - 3) / np.sqrt(
            (0.0025 / 0.15 ** 3) * 0.5 + (0.0003472 / 0.05833 ** 3) * 0.5)
        self.targ_h05 = [[0.5, 1, 1.5],
                         [(0 - 1.7) / np.sqrt(0.4), (0 - 1.7) / np.sqrt(0.4),
                          (x - 1.7) / np.sqrt(0.4)]]

    def test_filter_process_with_spiketrain_h05(self):
        st = neo.SpikeTrain(self.test_array, units='s', t_stop=2.1)
        target = self.targ_h05
        res = mft._filter_process(0.5 * pq.s, 0.5 * pq.s, st, 2.01 * pq.s,
                                  np.array([[0.5], [1.7], [0.4]]))
        assert_array_almost_equal(res[1], target[1], decimal=3)

        self.assertRaises(ValueError, mft._filter_process, 0.5, 0.5 * pq.s,
                          st, 2.01 * pq.s, np.array([[0.5], [1.7], [0.4]]))
        self.assertRaises(ValueError, mft._filter_process, 0.5 * pq.s, 0.5,
                          st, 2.01 * pq.s, np.array([[0.5], [1.7], [0.4]]))
        self.assertRaises(ValueError, mft._filter_process, 0.5 * pq.s,
                          0.5 * pq.s, self.test_array, 2.01 * pq.s,
                          np.array([[0.5], [1.7], [0.4]]))

    def test_filter_proces_with_quantities_h05(self):
        st = pq.Quantity(self.test_array, units='s')
        target = self.targ_h05
        res = mft._filter_process(0.5 * pq.s, 0.5 * pq.s, st, 2.01 * pq.s,
                                  np.array([[0.5], [1.7], [0.4]]))
        assert_array_almost_equal(res[0], target[0], decimal=3)

    def test_filter_proces_with_plain_array_h05(self):
        st = self.test_array
        target = self.targ_h05
        res = mft._filter_process(0.5 * pq.s, 0.5 * pq.s, st * pq.s,
                                  2.01 * pq.s, np.array([[0.5], [1.7], [0.4]]))
        self.assertNotIsInstance(res, pq.Quantity)
        assert_array_almost_equal(res, target, decimal=3)


class MultipleFilterAlgorithmTestCase(unittest.TestCase):
    def setUp(self):
        self.test_array = [1.1, 1.2, 1.4, 1.6, 1.7, 1.75, 1.8, 1.85, 1.9, 1.95]
        self.targ_h05_dt05 = [1.5 * pq.s]

        # to speed up the test, the following `test_param` and `test_quantile`
        # paramters have been calculated offline using the function:
        # empirical_parameters([10, 25, 50, 75, 100, 125, 150]*pq.s,700*pq.s,5,
        #                                                                10000)
        # the user should do the same, if the metohd has to be applied to
        # several spike trains of the same length `T` and with the same set of
        # window.
        self.test_param = np.array([[10.,
                                     25.,
                                     50.,
                                     75.,
                                     100.,
                                     125.,
                                     150.],
                                    [3.167,
                                     2.955,
                                     2.721,
                                     2.548,
                                     2.412,
                                     2.293,
                                     2.180],
                                    [0.150,
                                     0.185,
                                     0.224,
                                     0.249,
                                     0.269,
                                     0.288,
                                     0.301]])
        self.test_quantile = 2.75

    def test_MultipleFilterAlgorithm_with_spiketrain_h05(self):
        st = neo.SpikeTrain(self.test_array, units='s', t_stop=2.1)
        target = [self.targ_h05_dt05]
        res = mft.multiple_filter_test([0.5] * pq.s, st, 2.1 * pq.s, 5, 100,
                                       time_step=0.1 * pq.s)
        assert_array_almost_equal(res, target, decimal=9)

    def test_MultipleFilterAlgorithm_with_quantities_h05(self):
        st = pq.Quantity(self.test_array, units='s')
        target = [self.targ_h05_dt05]
        res = mft.multiple_filter_test([0.5] * pq.s, st, 2.1 * pq.s, 5, 100,
                                       time_step=0.5 * pq.s)
        self.assertNotIsInstance(res, pq.Quantity)
        assert_array_almost_equal(res, target, decimal=9)

    def test_MultipleFilterAlgorithm_with_longdata(self):

        def gamma_train(k, teta, tmax):
            x = np.random.gamma(k, teta, int(tmax * (k * teta) ** (-1) * 3))
            s = np.cumsum(x)
            idx = np.where(s < tmax)
            s = s[idx]  # gamma process
            return s

        def alternative_hypothesis(k1, teta1, c1, k2, teta2, c2, k3, teta3, c3,
                                   k4, teta4, T):
            s1 = gamma_train(k1, teta1, c1)
            s2 = gamma_train(k2, teta2, c2) + c1
            s3 = gamma_train(k3, teta3, c3) + c1 + c2
            s4 = gamma_train(k4, teta4, T) + c1 + c2 + c3
            return np.concatenate((s1, s2, s3, s4)), [s1[-1], s2[-1], s3[-1]]

        st = self.h1 = alternative_hypothesis(1, 1 / 4., 150, 2, 1 / 26., 30,
                                              1, 1 / 36., 320,
                                              2, 1 / 33., 200)[0]

        window_size = [10, 25, 50, 75, 100, 125, 150] * pq.s
        self.target_points = [150, 180, 500]
        target = self.target_points

        result = mft.multiple_filter_test(
            window_size,
            st * pq.s,
            700 * pq.s,
            5,
            10000,
            test_quantile=self.test_quantile,
            test_param=self.test_param,
            time_step=1 * pq.s)
        self.assertNotIsInstance(result, pq.Quantity)

        result_concatenated = []
        for i in result:
            result_concatenated = np.hstack([result_concatenated, i])
        result_concatenated = np.sort(result_concatenated)
        assert_allclose(result_concatenated[:3], target[:3], rtol=0,
                        atol=5)
        print('detected {0} cps: {1}'.format(len(result_concatenated),
                                             result_concatenated))


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
"""
Functions to measure the synchrony of several spike trains.


Synchrony Measures
------------------

.. autosummary::
    :toctree: _toctree/spike_train_synchrony/

    spike_contrast
    Synchrotool


:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
from __future__ import division, print_function, unicode_literals

import warnings
from collections import namedtuple
from copy import deepcopy

import neo
import numpy as np
import quantities as pq

from elephant.statistics import Complexity
from elephant.utils import is_time_quantity, check_same_units

SpikeContrastTrace = namedtuple("SpikeContrastTrace", (
    "contrast", "active_spiketrains", "synchrony", "bin_size"))


__all__ = [
    "SpikeContrastTrace",
    "spike_contrast",
    "Synchrotool"
]


def _get_theta_and_n_per_bin(spiketrains, t_start, t_stop, bin_size):
    """
    Calculates theta (amount of spikes per bin) and the amount of active spike
    trains per bin of one spike train.
    """
    bin_step = bin_size / 2
    edges = np.arange(t_start, t_stop + bin_step, bin_step)
    # Calculate histogram for every spike train
    histogram = np.vstack([
        _binning_half_overlap(st, edges=edges)
        for st in spiketrains
    ])
    # Amount of spikes per bin
    theta = histogram.sum(axis=0)
    # Amount of active spike trains per bin
    n_active_per_bin = np.count_nonzero(histogram, axis=0)

    return theta, n_active_per_bin


def _binning_half_overlap(spiketrain, edges):
    """
    Referring to [1] overlapping the bins creates a better result.
    """
    histogram, bin_edges = np.histogram(spiketrain, bins=edges)
    histogram = histogram[:-1] + histogram[1:]
    return histogram


def spike_contrast(spiketrains, t_start=None, t_stop=None,
                   min_bin=10 * pq.ms, bin_shrink_factor=0.9,
                   return_trace=False):
    """
    Calculates the synchrony of spike trains, according to
    :cite:`synchrony-Ciba18_136`. The spike trains can have different lengths.

    Original implementation by: Philipp Steigerwald [s160857@th-ab.de]

    Visualization is covered in
    :func:`viziphant.spike_train_synchrony.plot_spike_contrast`.

    Parameters
    ----------
    spiketrains : list of neo.SpikeTrain
        A list of input spike trains to calculate the synchrony from.
    t_start : pq.Quantity, optional
        The beginning of the spike train. If None, it's taken as the minimum
        value of `t_start`s of the input spike trains.
        Default: None
    t_stop : pq.Quantity, optional
        The end of the spike train. If None, it's taken as the maximum value
        of `t_stop` of the input spike trains.
        Default: None
    min_bin : pq.Quantity, optional
        Sets the minimum value for the `bin_min` that is calculated by the
        algorithm and defines the smallest bin size to compute the histogram
        of the input `spiketrains`.
        Default: 0.01 ms
    bin_shrink_factor : float, optional
        A multiplier to shrink the bin size on each iteration. The value must
        be in range `(0, 1)`.
        Default: 0.9
    return_trace : bool, optional
        If set to True, returns a history of spike-contrast synchrony, computed
        for a range of different bin sizes, alongside with the maximum value of
        the synchrony.
        Default: False

    Returns
    -------
    synchrony : float
        Returns the synchrony of the input spike trains.
    spike_contrast_trace : namedtuple
        If `return_trace` is set to True, a `SpikeContrastTrace` namedtuple is
        returned with the following attributes:

          `.contrast` - the average sum of differences of the number of spikes
          in subsuequent bins;

          `.active_spiketrains` - the average number of spikes per bin,
          weighted by the number of spike trains containing at least one spike
          inside the bin;

          `.synchrony` - the product of `contrast` and `active_spiketrains`;

          `.bin_size` - the X axis, a list of bin sizes that correspond to
          these traces.

    Raises
    ------
    ValueError
        If `bin_shrink_factor` is not in (0, 1) range.

        If the input spike trains constist of a single spiketrain.

        If all input spike trains contain no more than 1 spike.
    TypeError
        If the input spike trains is not a list of `neo.SpikeTrain` objects.

        If `t_start`, `t_stop`, or `min_bin` are not time quantities.

    Examples
    --------
    >>> import quantities as pq
    >>> from elephant.spike_train_generation import homogeneous_poisson_process
    >>> from elephant.spike_train_synchrony import spike_contrast
    >>> spiketrain_1 = homogeneous_poisson_process(rate=20*pq.Hz,
    ...     t_stop=1000*pq.ms)
    >>> spiketrain_2 = homogeneous_poisson_process(rate=20*pq.Hz,
    ...     t_stop=1000*pq.ms)
    >>> spike_contrast([spiketrain_1, spiketrain_2])
    0.4192546583850932

    """
    if not 0. < bin_shrink_factor < 1.:
        raise ValueError(f"'bin_shrink_factor' ({bin_shrink_factor}) must be "
                         "in range (0, 1).")
    if not len(spiketrains) > 1:
        raise ValueError("Spike contrast measure requires more than 1 input "
                         "spiketrain.")
    check_same_units(spiketrains, object_type=neo.SpikeTrain)
    if not is_time_quantity(t_start, t_stop, allow_none=True):
        raise TypeError("'t_start' and 't_stop' must be time quantities.")
    if not is_time_quantity(min_bin):
        raise TypeError("'min_bin' must be a time quantity.")

    if t_start is None:
        t_start = min(st.t_start for st in spiketrains)
    if t_stop is None:
        t_stop = max(st.t_stop for st in spiketrains)

    # convert everything to spiketrain units
    units = spiketrains[0].units
    spiketrains = [st.magnitude for st in spiketrains]
    t_start = t_start.rescale(units).item()
    t_stop = t_stop.rescale(units).item()
    min_bin = min_bin.rescale(units).item()

    spiketrains = [times[(times >= t_start) & (times <= t_stop)]
                   for times in spiketrains]

    n_spiketrains = len(spiketrains)
    n_spikes_total = sum(map(len, spiketrains))

    duration = t_stop - t_start
    bin_max = duration / 2

    try:
        isi_min = min(np.diff(st).min() for st in spiketrains if len(st) > 1)
    except TypeError:
        raise ValueError("All input spiketrains contain no more than 1 spike.")
    bin_min = max(isi_min / 2, min_bin)

    contrast_list = []
    active_spiketrains = []
    synchrony_curve = []

    # Set new time boundaries
    t_start = t_start - isi_min
    t_stop = t_stop + isi_min

    bin_sizes = []
    bin_size = bin_max
    while bin_size >= bin_min:
        bin_sizes.append(bin_size)
        # Calculate Theta and n
        theta_k, n_k = _get_theta_and_n_per_bin(spiketrains,
                                                t_start=t_start,
                                                t_stop=t_stop,
                                                bin_size=bin_size)

        # calculate synchrony_curve = contrast * active_st
        active_st = (np.sum(n_k * theta_k) / np.sum(theta_k) - 1) / (
                    n_spiketrains - 1)
        contrast = np.sum(np.abs(np.diff(theta_k))) / (2 * n_spikes_total)
        # Contrast: sum(|derivation|) / (2*#Spikes)
        synchrony = contrast * active_st

        contrast_list.append(contrast)
        active_spiketrains.append(active_st)
        synchrony_curve.append(synchrony)

        # New bin size
        bin_size *= bin_shrink_factor

    # Sync value is maximum of the cost function C
    synchrony = max(synchrony_curve)

    if return_trace:
        spike_contrast_trace = SpikeContrastTrace(
            contrast=contrast_list,
            active_spiketrains=active_spiketrains,
            synchrony=synchrony_curve,
            bin_size=bin_sizes * units,
        )
        return synchrony, spike_contrast_trace

    return synchrony


class Synchrotool(Complexity):
    """
    Tool class to find, remove and/or annotate the presence of synchronous
    spiking events across multiple spike trains.

    The complexity is used to characterize synchronous events within the same
    spike train and across different spike trains in the `spiketrains` list.
    This way synchronous events can be found both in multi-unit and
    single-unit spike trains.

    This class inherits from :class:`elephant.statistics.Complexity`, see its
    documentation for more details and input parameters description.

    See also
    --------
    elephant.statistics.Complexity

    """

    def __init__(self, spiketrains,
                 sampling_rate,
                 bin_size=None,
                 binary=True,
                 spread=0,
                 tolerance=1e-8):

        self.annotated = False

        super(Synchrotool, self).__init__(spiketrains=spiketrains,
                                          bin_size=bin_size,
                                          sampling_rate=sampling_rate,
                                          binary=binary,
                                          spread=spread,
                                          tolerance=tolerance)

    def delete_synchrofacts(self, threshold, in_place=False, mode='delete'):
        """
        Delete or extract synchronous spiking events.

        Parameters
        ----------
        threshold : int
            Threshold value for the deletion of spikes engaged in synchronous
            activity.
              * `deletion_threshold >= 2` leads to all spikes with a larger or
                equal complexity value to be deleted/extracted.
              * `deletion_threshold <= 1` leads to a ValueError, since this
              would delete/extract all spikes and there are definitely more
              efficient ways of doing so.
        in_place : bool, optional
            Determines whether the modification are made in place
            on ``self.input_spiketrains``.
            Default: False
        mode : {'delete', 'extract'}, optional
            Inversion of the mask for deletion of synchronous events.
              * ``'delete'`` leads to the deletion of all spikes with
                complexity >= `threshold`,
                i.e. deletes synchronous spikes.
              * ``'extract'`` leads to the deletion of all spikes with
                complexity < `threshold`, i.e. extracts synchronous spikes.
            Default: 'delete'

        Raises
        ------
        ValueError
            If `mode` is not one in {'delete', 'extract'}.

            If `threshold <= 1`.

        Returns
        -------
        list of neo.SpikeTrain
            List of spiketrains where the spikes with
            ``complexity >= threshold`` have been deleted/extracted.
              * If ``in_place`` is True, the returned list is the same as
                ``self.input_spiketrains``.
              * If ``in_place`` is False, the returned list is a deepcopy of
                ``self.input_spiketrains``.

        """

        if not self.annotated:
            self.annotate_synchrofacts()

        if mode not in ['delete', 'extract']:
            raise ValueError(f"Invalid mode '{mode}'. Valid modes are: "
                             f"'delete', 'extract'")

        if threshold <= 1:
            raise ValueError('A deletion threshold <= 1 would result '
                             'in the deletion of all spikes.')

        if in_place:
            spiketrain_list = self.input_spiketrains
        else:
            spiketrain_list = deepcopy(self.input_spiketrains)

        for idx, st in enumerate(spiketrain_list):
            mask = st.array_annotations['complexity'] < threshold
            if mode == 'extract':
                mask = np.invert(mask)
            new_st = st[mask]
            if in_place and st.segment is not None:
                segment = st.segment

                try:
                    # replace link to spiketrain in segment
                    new_index = self._get_spiketrain_index(
                        segment.spiketrains, st)
                    # Todo: Simplify following lines once Neo SpikeTrainList
                    # implments indexed assignment of entries (i.e., stl[i]=st)
                    spiketrainlist = list(segment.spiketrains)
                    spiketrainlist[new_index] = new_st
                    segment.spiketrains = spiketrainlist
                except ValueError:
                    # st is not in this segment even though it points to it
                    warnings.warn(f"The SpikeTrain at index {idx} of the "
                                  "input list spiketrains has a "
                                  "unidirectional uplink to a segment in "
                                  "whose segment.spiketrains list it does not "
                                  "appear. Only the spiketrains in the input "
                                  "list will be replaced. You can suppress "
                                  "this warning by setting "
                                  "spiketrain.segment=None for the input "
                                  "spiketrains.")

                block = segment.block
                if block is not None:
                    # replace link to spiketrain in groups
                    for group in block.groups:
                        try:
                            idx = self._get_spiketrain_index(
                                group.spiketrains,
                                st)
                        except ValueError:
                            # st is not in this group, move to next group
                            continue

                        # st found in group, replace with new_st
                        group.spiketrains[idx] = new_st
            spiketrain_list[idx] = new_st

        return spiketrain_list

    def annotate_synchrofacts(self):
        """
        Annotate the complexity of each spike in the
        ``self.epoch.array_annotations`` *in-place*.
        """
        epoch_complexities = self.epoch.array_annotations['complexity']
        right_edges = (
            self.epoch.times.magnitude.flatten()
            + self.epoch.durations.rescale(
                self.epoch.times.units).magnitude.flatten()
        )

        for idx, st in enumerate(self.input_spiketrains):

            # all indices of spikes that are within the half-open intervals
            # defined by the boundaries
            # note that every second entry in boundaries is an upper boundary
            spike_to_epoch_idx = np.searchsorted(
                right_edges,
                st.times.rescale(self.epoch.times.units).magnitude.flatten())
            complexity_per_spike = epoch_complexities[spike_to_epoch_idx]

            st.array_annotate(complexity=complexity_per_spike)

        self.annotated = True

    def _get_spiketrain_index(self, spiketrain_list, spiketrain):
        for index, item in enumerate(spiketrain_list):
            if item is spiketrain:
                return index
        raise ValueError("Spiketrain is not found in the list")



# -*- coding: utf-8 -*-
"""
Unitary Event (UE) analysis is a statistical method to analyze in a time
resolved manner excess spike correlation between simultaneously recorded
neurons by comparing the empirical spike coincidences (precision of a few ms)
to the expected number based on the firing rates of the neurons
(see :cite:`unitary_event_analysis-Gruen99_67`).

Background
----------

It has been proposed that cortical neurons organize dynamically into functional
groups (“cell assemblies”) by the temporal structure of their joint spiking
activity. The Unitary Events analysis method detects conspicuous patterns of
synchronous spike activity among simultaneously recorded single neurons. The
statistical significance of a pattern is evaluated by comparing the empirical
number of occurrences to the number expected given the firing rates of the
neurons. Key elements of the method are the proper formulation of the null
hypothesis and the derivation of the corresponding count distribution of
synchronous spike events used in the significance test. The analysis is
performed in a sliding window manner and yields a time-resolved measure of
significant spike synchrony. For further reading, see
:cite:`unitary_event_analysis-Riehle97_1950,unitary_event_analysis-Gruen02_43,\
unitary_event_analysis-Gruen02_81,unitary_event_analysis-Gruen03,\
unitary_event_analysis-Gruen09_1126,unitary_event_analysis-Gruen99_67`.


Tutorial
--------

:doc:`View tutorial <../tutorials/unitary_event_analysis>`

Run tutorial interactively:

.. image:: https://mybinder.org/badge.svg
   :target: https://mybinder.org/v2/gh/NeuralEnsemble/elephant/master
            ?filepath=doc/tutorials/unitary_event_analysis.ipynb


Functions overview
------------------

.. autosummary::
    :toctree: _toctree/unitary_event_analysis/

    jointJ_window_analysis

:copyright: Copyright 2014-2022 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import warnings
from collections import defaultdict

import neo
import numpy as np
import quantities as pq
import scipy

import elephant.conversion as conv
from elephant.utils import is_binary, deprecated_alias

__all__ = [
    "hash_from_pattern",
    "inverse_hash_from_pattern",
    "n_emp_mat",
    "n_emp_mat_sum_trial",
    "n_exp_mat",
    "n_exp_mat_sum_trial",
    "gen_pval_anal",
    "jointJ",
    "jointJ_window_analysis"
]


def hash_from_pattern(m, base=2):
    """
    Calculate for a spike pattern or a matrix of spike patterns
    (provide each pattern as a column) composed of N neurons a
    unique number.


    Parameters
    ----------
    m: np.ndarray or list
        2-dim ndarray
        spike patterns represented as a binary matrix (i.e., matrix of 0's and
        1's).
        Rows and columns correspond to patterns and neurons, respectively.
    base: integer
        The base for hashes calculation.
        Default: 2

    Returns
    -------
    np.ndarray
        An array containing the hash values of each pattern,
        shape: (number of patterns).

    Raises
    ------
    ValueError
        If matrix `m` has wrong orientation.

    Examples
    --------
    With `base=2`, the hash of `[0, 1, 1]` is `0*2^2 + 1*2^1 + 1*2^0 = 3`.

    >>> import numpy as np
    >>> hash_from_pattern([0, 1, 1])
    3

    >>> import numpy as np
    >>> m = np.array([[0, 1, 0, 0, 1, 1, 0, 1],
    ...               [0, 0, 1, 0, 1, 0, 1, 1],
    ...               [0, 0, 0, 1, 0, 1, 1, 1]])

    >>> hash_from_pattern(m)
    array([0, 4, 2, 1, 6, 5, 3, 7])

    """
    m = np.asarray(m)
    n_neurons = m.shape[0]

    # check the entries of the matrix
    if not is_binary(m):
        raise ValueError('Patterns should be binary: 0 or 1')

    # generate the representation
    # don't use numpy - it's upperbounded by int64
    powers = [base ** x for x in range(n_neurons)][::-1]

    # calculate the binary number by use of scalar product
    return np.dot(powers, m)


def inverse_hash_from_pattern(h, N, base=2):
    """
    Calculate the binary spike patterns (matrix) from hash values `h`.

    Parameters
    ----------
    h: list of int
        Array-like of integer hash values of length of the number of patterns.
    N: integer
        The number of neurons.
    base: integer
        The base, used to generate the hash values.
        Default: 2

    Returns
    -------
    m: (N, P) np.ndarray
       A matrix of shape: (N, number of patterns)

    Raises
    ------
    ValueError
        If the hash is not compatible with the number of neurons.
        The hash value should not be larger than the largest
        possible hash number with the given number of neurons
        (e.g. for N = 2, max(hash) = 2^1 + 2^0 = 3, or for N = 4,
        max(hash) = 2^3 + 2^2 + 2^1 + 2^0 = 15).

    Examples
    ---------
    >>> import numpy as np
    >>> h = np.array([3, 7])
    >>> N = 4
    >>> inverse_hash_from_pattern(h, N)
    array([[1, 1],
        [1, 1],
        [0, 1],
        [0, 0]])

    """
    h = np.asarray(h)  # this will cast to object type if h > int64
    if not all(isinstance(v, int) for v in h.tolist()):
        # .tolist() is necessary because np.int[64] is not int
        raise ValueError("hash values should be integers")

    # check if the hash values are not greater than the greatest possible
    # value for N neurons with the given base
    powers = np.array([base ** x for x in range(N)])[::-1]
    if any(h > sum(powers)):
        raise ValueError(f"hash value {h} is not compatible with the number "
                         f"of neurons {N}")
    m = h // np.expand_dims(powers, axis=1)
    m %= base  # m is a binary matrix now
    m = m.astype(int)  # convert object to int if the hash was > int64
    return m


def n_emp_mat(mat, pattern_hash, base=2):
    """
    Count the occurrences of spike coincidence patterns in the given spike
    trains.

    Parameters
    ----------
    mat : (N, M) np.ndarray
        Binned spike trains of N neurons. Rows and columns correspond
        to neurons and temporal bins, respectively.
    pattern_hash: list of int
        List of hash values, representing the spike coincidence patterns
        of which occurrences are counted.
    base: integer
        The base, used to generate the hash values.
        Default: 2

    Returns
    -------
    N_emp: np.ndarray
        The number of occurrences of the given patterns in the given
        spiketrains.
    indices: list of list
        List of lists of int.
        Indices indexing the bins where the given spike patterns are found
        in `mat`. Same length as `pattern_hash`.
        `indices[i] = N_emp[i] = pattern_hash[i]`

    Raises
    ------
    ValueError
        If `mat` is not a binary matrix.

    Examples
    --------
    >>> mat = np.array([[1, 0, 0, 1, 1],
    ...                 [1, 0, 0, 1, 0]])
    >>> pattern_hash = np.array([1,3])
    >>> n_emp, n_emp_indices = n_emp_mat(mat, pattern_hash)
    >>> print(n_emp)
    [ 0.  2.]
    >>> print(n_emp_indices)
    [array([]), array([0, 3])]

    """
    # check if the mat is zero-one matrix
    if not is_binary(mat):
        raise ValueError("entries of mat should be either one or zero")
    h = hash_from_pattern(mat, base=base)
    N_emp = np.zeros(len(pattern_hash))
    indices = []
    for idx_ph, ph in enumerate(pattern_hash):
        indices_tmp = np.where(h == ph)[0]
        indices.append(indices_tmp)
        N_emp[idx_ph] = len(indices_tmp)
    return N_emp, indices


def n_emp_mat_sum_trial(mat, pattern_hash):
    """
    Calculate empirical number of observed patterns, summed across trials.

    Parameters
    ----------
    mat: np.ndarray
        Binned spike trains are represented as a binary matrix (i.e., matrix of
        0's and 1's), segmented into trials. Trials should contain an identical
        number of neurons and an identical number of time bins.
         the entries are zero or one
         0-axis --> trials
         1-axis --> neurons
         2-axis --> time bins
    pattern_hash: list of int
         Array of hash values of length of the number of patterns.

    Returns
    -------
    N_emp: np.ndarray
        The number of occurences of the given spike patterns in the given spike
        trains, summed across trials. Same length as `pattern_hash`.
    idx_trials: list of int
        List of indices of `mat` for each trial in which the specific pattern
        has been observed.
        0-axis --> trial
        1-axis --> list of indices for the chosen trial per entry of
                    `pattern_hash`

    Raises
    ------
    ValueError
        If `mat` has the wrong orientation.
        If `mat` is not a binary matrix.

    Examples
    ---------
    >>> mat = np.array([[[1, 1, 1, 1, 0],
    ...                  [0, 1, 1, 1, 0],
    ...                  [0, 1, 1, 0, 1]],
    ...                 [[1, 1, 1, 1, 1],
    ...                  [0, 1, 1, 1, 1],
    ...                  [1, 1, 0, 1, 0]]])
    >>> pattern_hash = np.array([4,6])
    >>> n_emp_sum_trial, n_emp_sum_trial_idx = \
    ...                   n_emp_mat_sum_trial(mat, pattern_hash)
    >>> n_emp_sum_trial
    array([ 1.,  3.])
    >>> n_emp_sum_trial_idx
    [[array([0]), array([3])], [array([], dtype=int64), array([2, 4])]]

    """
    num_patt = len(pattern_hash)
    N_emp = np.zeros(num_patt)

    idx_trials = []
    # check if the mat is zero-one matrix
    if not is_binary(mat):
        raise ValueError("entries of mat should be either one or zero")

    for mat_tr in mat:
        N_emp_tmp, indices_tmp = n_emp_mat(mat_tr, pattern_hash, base=2)
        idx_trials.append(indices_tmp)
        N_emp += N_emp_tmp

    return N_emp, idx_trials


def _n_exp_mat_analytic(mat, pattern_hash):
    """
    Calculates the expected joint probability for each spike pattern
    analytically.
    """
    marg_prob = np.mean(mat, 1, dtype=float)
    # marg_prob needs to be a column vector, so we
    # build a two dimensional array with 1 column
    # and len(marg_prob) rows
    marg_prob = np.expand_dims(marg_prob, axis=1)
    n_neurons = mat.shape[0]
    m = inverse_hash_from_pattern(pattern_hash, n_neurons)
    nrep = m.shape[1]
    # multipyling the marginal probability of neurons with regard to the
    # pattern
    pmat = np.multiply(m, np.tile(marg_prob, (1, nrep))) \
        + np.multiply(1 - m, np.tile(1 - marg_prob, (1, nrep)))
    return np.prod(pmat, axis=0) * float(mat.shape[1])


def _n_exp_mat_surrogate(mat, pattern_hash, n_surrogates=1):
    """
    Calculates the expected joint probability for each spike pattern with spike
    time randomization surrogate
    """
    if len(pattern_hash) > 1:
        raise ValueError('surrogate method works only for one pattern!')
    N_exp_array = np.zeros(n_surrogates)
    for rz_idx, rz in enumerate(np.arange(n_surrogates)):
        # row-wise shuffling all elements of zero-one matrix
        mat_surr = np.copy(mat)
        for row in mat_surr:
            np.random.shuffle(row)
        N_exp_array[rz_idx] = n_emp_mat(mat_surr, pattern_hash)[0][0]
    return N_exp_array


def n_exp_mat(mat, pattern_hash, method='analytic', n_surrogates=1):
    """
    Calculates the expected joint probability for each spike pattern.

    Parameters
    ----------
    mat: np.ndarray
         The entries are in the range [0, 1].
         The only possibility when the entries are floating point values is
         when the `mat` is calculated with the flag `analytic_TrialAverage`
         in `n_exp_mat_sum_trial()`.
         Otherwise, the entries are binary.
         0-axis --> neurons
         1-axis --> time bins
    pattern_hash: list of int
         List of hash values, length: number of patterns
    method: {'analytic', 'surr'}, optional
         The method with which the expectation is calculated.
         'analytic' -- > analytically
         'surr' -- > with surrogates (spike time randomization)
         Default: 'analytic'
    n_surrogates: int
         number of surrogates for constructing the distribution of expected
         joint probability.
         Default: 1 and this number is needed only when method = 'surr'

    Returns
    -------
    np.ndarray
        if method is 'analytic':
            An array containing the expected joint probability of each pattern,
            shape: (number of patterns,)
        if method is 'surr':
            0-axis --> different realizations, length = number of surrogates
            1-axis --> patterns

    Raises
    ------
    ValueError
        If `mat` has the wrong orientation.

    Examples
    --------
    >>> mat = np.array([[1, 1, 1, 1],
    ...                 [0, 1, 0, 1],
    ...                 [0, 0, 1, 0]])
    >>> pattern_hash = np.array([5,6])
    >>> n_exp_anal = n_exp_mat(mat, pattern_hash, method='analytic')
    >>> n_exp_anal
    [ 0.5 1.5 ]
    >>> n_exp_surr = n_exp_mat(mat, pattern_hash, method='surr',
    ...                        n_surrogates=5000)
    >>> print(n_exp_surr)
    [[ 1.  1.]
     [ 2.  0.]
     [ 2.  0.]
     ...,
     [ 2.  0.]
     [ 2.  0.]
     [ 1.  1.]]

    """
    # check if the mat is in the range [0, 1]
    if not np.all((mat >= 0) & (mat <= 1)):
        raise ValueError("entries of mat should be in range [0, 1]")

    if method == 'analytic':
        return _n_exp_mat_analytic(mat, pattern_hash)
    if method == 'surr':
        return _n_exp_mat_surrogate(mat, pattern_hash,
                                    n_surrogates=n_surrogates)


def n_exp_mat_sum_trial(mat, pattern_hash, method='analytic_TrialByTrial',
                        n_surrogates=1):
    """
    Calculates the expected joint probability for each spike pattern sum over
    trials.

    Parameters
    ----------
    mat: np.ndarray
        Binned spike trains represented as a binary matrix (i.e., matrix of
        0's and 1's), segmented into trials. Trials should contain an identical
        number of neurons and an identical number of time bins.
        The entries of mat should be a list of a list where 0-axis is trials
        and 1-axis is neurons.
         0-axis --> trials
         1-axis --> neurons
         2-axis --> time bins
    pattern_hash: list of int
         List of hash values, length: number of patterns
    method: str
         method with which the unitary events whould be computed
         'analytic_TrialByTrial' -- > calculate the expectency
         (analytically) on each trial, then sum over all trials.
         'analytic_TrialAverage' -- > calculate the expectency
         by averaging over trials.
         (cf. Gruen et al. 2003)
         'surrogate_TrialByTrial' -- > calculate the distribution
         of expected coincidences by spike time randomzation in
         each trial and sum over trials.
         Default: 'analytic_trialByTrial'.
    n_surrogates: int, optional
         The number of surrogate to be used.
         Default: 1

    Returns
    -------
    n_exp: np.ndarray
         An array containing the expected joint probability of
         each pattern summed over trials,shape: (number of patterns,)

    Raises
    ------
    ValueError
        If `method` is not one of the specified above.

    Examples
    --------
    >>> mat = np.array([[[1, 1, 1, 1, 0],
    ...                  [0, 1, 1, 1, 0],
    ...                  [0, 1, 1, 0, 1]],
    ...                 [[1, 1, 1, 1, 1],
    ...                  [0, 1, 1, 1, 1],
    ...                  [1, 1, 0, 1, 0]]])

    >>> pattern_hash = np.array([5,6])
    >>> n_exp_anal = n_exp_mat_sum_trial(mat, pattern_hash)
    >>> print(n_exp_anal)
        array([ 1.56,  2.56])
    """
    if method == 'analytic_TrialByTrial':
        n_exp = np.zeros(len(pattern_hash))
        for mat_tr in mat:
            n_exp += n_exp_mat(mat_tr, pattern_hash,
                               method='analytic')
    elif method == 'analytic_TrialAverage':
        n_exp = n_exp_mat(
            np.mean(mat, axis=0), pattern_hash,
            method='analytic') * mat.shape[0]
    elif method == 'surrogate_TrialByTrial':
        n_exp = np.zeros(n_surrogates)
        for mat_tr in mat:
            n_exp += n_exp_mat(mat_tr, pattern_hash,
                               method='surr', n_surrogates=n_surrogates)
    else:
        raise ValueError(
            "The method only works on the zero_one matrix at the moment")
    return n_exp


def gen_pval_anal(mat, pattern_hash, method='analytic_TrialByTrial',
                  n_surrogates=1):
    """
    Compute the expected coincidences and a function to calculate the
    p-value for the given empirical coincidences.

    This function generates a poisson distribution with the expected
    value calculated by `mat`. It returns a function that gets
    the empirical coincidences, `n_emp`, and calculates a p-value
    as the area under the poisson distribution from `n_emp` to infinity.

    Parameters
    ----------
    mat: np.ndarray
        Binned spike trains represented as a binary matrix (i.e., matrix of
        0's and 1's), segmented into trials. Trials should contain an identical
        number of neurons and an identical number of time bins.
        The entries of mat should be a list of a list where 0-axis is trials
        and 1-axis is neurons.
         0-axis --> trials
         1-axis --> neurons
         2-axis --> time bins
    pattern_hash: list of int
         List of hash values, length: number of patterns
    method: string
         method with which the unitary events whould be computed
         'analytic_TrialByTrial' -- > calculate the expectency
         (analytically) on each trial, then sum over all trials.
         ''analytic_TrialAverage' -- > calculate the expectency
         by averaging over trials.
         Default: 'analytic_trialByTrial'
         (cf. Gruen et al. 2003)
    n_surrogates: integer, optional
         number of surrogate to be used
         Default: 1

    Returns
    --------
    pval_anal: callable
         The function that calculates the p-value for the given empirical
         coincidences.
    n_exp: list
        List of expected coincidences.

    Raises
    ------
    ValueError
        If `method` is not one of the specified above.

    Examples
    --------
    >>> mat = np.array([[[1, 1, 1, 1, 0],
    ...                  [0, 1, 1, 1, 0],
    ...                  [0, 1, 1, 0, 1]],
    ...                 [[1, 1, 1, 1, 1],
    ...                  [0, 1, 1, 1, 1],
    ...                  [1, 1, 0, 1, 0]]])
    >>> pattern_hash = np.array([5, 6])
    >>> pval_anal, n_exp = gen_pval_anal(mat, pattern_hash)
    >>> n_exp
    array([ 1.56,  2.56])

    """
    if method == 'analytic_TrialByTrial' or method == 'analytic_TrialAverage':
        n_exp = n_exp_mat_sum_trial(mat, pattern_hash, method=method)

        def pval(n_emp):
            p = 1. - scipy.special.gammaincc(n_emp, n_exp)
            return p
    elif method == 'surrogate_TrialByTrial':
        n_exp = n_exp_mat_sum_trial(
            mat, pattern_hash, method=method, n_surrogates=n_surrogates)

        def pval(n_emp):
            hist = np.bincount(np.int64(n_exp))
            exp_dist = hist / float(np.sum(hist))
            if len(n_emp) > 1:
                raise ValueError('In surrogate method the p_value can be'
                                 'calculated only for one pattern!')
            return np.sum(exp_dist[int(n_emp[0]):])
    else:
        raise ValueError("Method is not allowed: {method}".format(
            method=method))

    return pval, n_exp


def jointJ(p_val):
    """
    Surprise measurement.

    Logarithmic transformation of joint-p-value into surprise measure
    for better visualization as the highly significant events are
    indicated by very low joint-p-values.

    Parameters
    ----------
    p_val : float or list of float
        List of p-values of statistical tests for different pattern.

    Returns
    -------
    Js: list of float
        List of surprise measures.

    Examples
    --------
    >>> p_val = np.array([0.31271072,  0.01175031])
    >>> jointJ(p_val)
    array([0.3419968 ,  1.92481736])

    """
    p_arr = np.asarray(p_val)
    with np.errstate(divide='ignore'):
        # Ignore 'Division by zero' warning which happens when p_arr is 1.0 or
        # 0.0 (no spikes).
        Js = np.log10(1 - p_arr) - np.log10(p_arr)
    return Js


def _rate_mat_avg_trial(mat):
    """
    Calculates the average firing rate of each neurons across trials.
    """
    n_trials, n_neurons, n_bins = np.shape(mat)
    psth = np.zeros(n_neurons, dtype=np.float32)
    for tr, mat_tr in enumerate(mat):
        psth += np.sum(mat_tr, axis=1)
    return psth / (n_bins * n_trials)


def _bintime(t, bin_size):
    """
    Change the real time to `bin_size` units.
    """
    t_dl = t.rescale('ms').magnitude
    bin_size_dl = bin_size.rescale('ms').item()
    return np.floor(np.array(t_dl) / bin_size_dl).astype(int)


@deprecated_alias(winsize='win_size', winstep='win_step')
def _winpos(t_start, t_stop, win_size, win_step, position='left-edge'):
    """
    Calculate the position of the analysis window.
    """
    t_start_dl = t_start.rescale('ms').item()
    t_stop_dl = t_stop.rescale('ms').item()
    winsize_dl = win_size.rescale('ms').item()
    winstep_dl = win_step.rescale('ms').item()

    # left side of the window time
    if position == 'left-edge':
        ts_winpos = np.arange(
            t_start_dl, t_stop_dl - winsize_dl + winstep_dl,
            winstep_dl) * pq.ms
    else:
        raise ValueError(
            'the current version only returns left-edge of the window')
    return ts_winpos


def _UE(mat, pattern_hash, method='analytic_TrialByTrial', n_surrogates=1):
    """
    Return the default results of unitary events analysis
    (Surprise, empirical coincidences and index of where it happened
    in the given mat, n_exp and average rate of neurons)
    """
    rate_avg = _rate_mat_avg_trial(mat)
    n_emp, indices = n_emp_mat_sum_trial(mat, pattern_hash)
    if method == 'surrogate_TrialByTrial':
        dist_exp, n_exp = gen_pval_anal(
            mat, pattern_hash, method, n_surrogates=n_surrogates)
        n_exp = np.mean(n_exp)
    elif method == 'analytic_TrialByTrial' or \
            method == 'analytic_TrialAverage':
        dist_exp, n_exp = gen_pval_anal(mat, pattern_hash, method)
    pval = dist_exp(n_emp)
    Js = jointJ(pval)
    return Js, rate_avg, n_exp, n_emp, indices


@deprecated_alias(data='spiketrains', binsize='bin_size', winsize='win_size',
                  winstep='win_step', n_surr='n_surrogates')
def jointJ_window_analysis(spiketrains, bin_size=5 * pq.ms,
                           win_size=100 * pq.ms, win_step=5 * pq.ms,
                           pattern_hash=None, method='analytic_TrialByTrial',
                           t_start=None, t_stop=None, binary=True,
                           n_surrogates=100):
    """
    Calculates the joint surprise in a sliding window fashion.

    Implementation is based on :cite:`unitary_event_analysis-Gruen99_67`.

    Parameters
    ----------
    spiketrains : list
        A list of spike trains (`neo.SpikeTrain` objects) in different trials:
          * 0-axis --> Trials

          * 1-axis --> Neurons

          * 2-axis --> Spike times
    bin_size : pq.Quantity, optional
        The size of bins for discretizing spike trains.
        Default: 5 ms
    win_size : pq.Quantity, optional
        The size of the window of analysis.
        Default: 100 ms
    win_step : pq.Quantity, optional
        The size of the window step.
        Default: 5 ms
    pattern_hash : int or list of int or None, optional
        A list of interested patterns in hash values (see `hash_from_pattern`
        and `inverse_hash_from_pattern` functions). If None, all neurons
        are participated.
        Default: None
    method : str, optional
        The method with which to compute the unitary events:
          * 'analytic_TrialByTrial': calculate the analytical expectancy
            on each trial, then sum over all trials;

          * 'analytic_TrialAverage': calculate the expectancy by averaging over
            trials (cf. Gruen et al. 2003);

          * 'surrogate_TrialByTrial': calculate the distribution of expected
            coincidences by spike time randomization in each trial and sum over
            trials.
        Default: 'analytic_trialByTrial'
    t_start, t_stop : float or pq.Quantity, optional
        The start and stop times to use for the time points.
        If not specified, retrieved from the `t_start` and `t_stop` attributes
        of the input spiketrains.
    binary : bool, optional
        Binarize the binned spike train objects (True) or not. Only the binary
        matrices are supported at the moment.
        Default: True
    n_surrogates : int, optional
        The number of surrogates to be used.
        Default: 100

    Returns
    -------
    dict
        The values of the following keys have the shape of

          * different window --> 0-axis
          * different pattern hash --> 1-axis

        'Js': list of float
          JointSurprise of different given patterns within each window.
        'indices': list of list of int
          A list of indices of pattern within each window.
        'n_emp': list of int
          The empirical number of each observed pattern.
        'n_exp': list of float
          The expected number of each pattern.
        'rate_avg': list of float
          The average firing rate of each neuron.

        Additionally, 'input_parameters' key stores the input parameters.

    Raises
    ------
    ValueError
        If `data` is not in the format, specified above.
    NotImplementedError
        If `binary` is not True. The method works only with binary matrices at
        the moment.

    Warns
    -----
    UserWarning
        The ratio between `winsize` or `winstep` and `bin_size` is not an
        integer.

    """
    if not isinstance(spiketrains[0][0], neo.SpikeTrain):
        raise ValueError(
            "structure of the data is not correct: 0-axis should be trials, "
            "1-axis units and 2-axis neo spike trains")

    if t_start is None:
        t_start = spiketrains[0][0].t_start
    if t_stop is None:
        t_stop = spiketrains[0][0].t_stop

    n_trials = len(spiketrains)
    n_neurons = len(spiketrains[0])
    if pattern_hash is None:
        pattern = [1] * n_neurons
        pattern_hash = hash_from_pattern(pattern)
    if np.issubdtype(type(pattern_hash), np.integer):
        pattern_hash = [int(pattern_hash)]

    # position of all windows (left edges)
    t_winpos = _winpos(t_start, t_stop, win_size, win_step,
                       position='left-edge')
    t_winpos_bintime = _bintime(t_winpos, bin_size)

    winsize_bintime = _bintime(win_size, bin_size)
    winstep_bintime = _bintime(win_step, bin_size)

    if winsize_bintime * bin_size != win_size:
        warnings.warn(f"The ratio between the win_size ({win_size}) and the "
                      f"bin_size ({bin_size}) is not an integer")

    if winstep_bintime * bin_size != win_step:
        warnings.warn(f"The ratio between the win_step ({win_step}) and the "
                      f"bin_size ({bin_size}) is not an integer")

    input_parameters = dict(pattern_hash=pattern_hash, bin_size=bin_size,
                            win_size=win_size, win_step=win_step,
                            method=method, t_start=t_start, t_stop=t_stop,
                            n_surrogates=n_surrogates)

    n_bins = int(((t_stop - t_start) / bin_size).simplified.item())

    mat_tr_unit_spt = np.zeros((len(spiketrains), n_neurons, n_bins),
                               dtype=np.int32)
    for trial, sts in enumerate(spiketrains):
        bs = conv.BinnedSpikeTrain(list(sts), t_start=t_start, t_stop=t_stop,
                                   bin_size=bin_size)
        if not binary:
            raise NotImplementedError(
                "The method works only with binary matrices at the moment")
        mat_tr_unit_spt[trial] = bs.to_bool_array()

    n_windows = len(t_winpos)
    n_hashes = len(pattern_hash)
    Js_win, n_exp_win, n_emp_win = np.zeros((3, n_windows, n_hashes),
                                            dtype=np.float32)
    rate_avg = np.zeros((n_windows, n_hashes, n_neurons), dtype=np.float32)
    indices_win = defaultdict(list)

    for i, win_pos in enumerate(t_winpos_bintime):
        mat_win = mat_tr_unit_spt[:, :, win_pos:win_pos + winsize_bintime]
        Js_win[i], rate_avg[i], n_exp_win[i], n_emp_win[
            i], indices_lst = _UE(mat_win, pattern_hash=pattern_hash,
                                  method=method, n_surrogates=n_surrogates)
        for j in range(n_trials):
            if len(indices_lst[j][0]) > 0:
                indices_win[f"trial{j}"].append(indices_lst[j][0] + win_pos)
    for key in indices_win.keys():
        indices_win[key] = np.hstack(indices_win[key])
    return {'Js': Js_win, 'indices': indices_win, 'n_emp': n_emp_win,
            'n_exp': n_exp_win, 'rate_avg': rate_avg / bin_size,
            'input_parameters': input_parameters}
